import collections
import math
from typing import Dict, List, Optional

import torch
from torch import Tensor, nn

from fusion_bench.utils.type import StateDictType


def compute_svd_dict(task_vectors, config):
    """
    Computes the Singular Value Decomposition (SVD) for each task vector in the provided datasets and stores the results in a dictionary.

    Args:
        task_vectors (list): A list of task vector objects, where each task vector contains a dictionary of matrices to be decomposed.
        config (object): Configuration object containing the list of datasets under the attribute `DATASETS`.

    Returns:
        dict: A dictionary where each key is a dataset name and the value is another dictionary containing the SVD components ('u', 's', 'v') for each matrix in the task vector.
        If a matrix is not 2-dimensional or contains 'text_projection' in its key, it is stored under the key 'dim1' without decomposition.
    """
    sv_reduction = 1 / len(config.DATASETS)
    with torch.no_grad():
        svd_dict = {}
        for i, (task_vector, dataset) in enumerate(zip(task_vectors, config.DATASETS)):
            svd_dict[dataset] = {}
            print(f"Computing SVD for {dataset}...")
            for key in task_vector.vector:
                svd_dict[dataset][key] = {}
                if (
                    len(task_vector.vector[key].shape) == 2
                    and "text_projection" not in key
                ):
                    u, s, v = torch.linalg.svd(
                        task_vector.vector[key], full_matrices=False
                    )
                    reduced_index_s = int(s.shape[0] * sv_reduction)

                    temp_u = torch.zeros_like(u)
                    # select only the first reduced_index_s columns of u and place them
                    temp_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                        :, :reduced_index_s
                    ]
                    svd_dict[dataset][key]["u"] = temp_u

                    temp_s = torch.zeros_like(s)
                    temp_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                        :reduced_index_s
                    ]

                    svd_dict[dataset][key]["s"] = temp_s  # s_reduced

                    # select only the first reduced_index_s rows of v and place them
                    temp_v = torch.zeros_like(v)
                    temp_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                        :reduced_index_s, :
                    ]

                    svd_dict[dataset][key]["v"] = temp_v

                    # temp_mat = temp_u @ torch.diag_embed(temp_s) @ temp_v
                else:
                    svd_dict[dataset][key]["dim1"] = task_vector.vector[key]
    return svd_dict


def sum_svd_dict(svd_dict, config):
    """
    Sums the Singular Value Decomposition (SVD) components from multiple datasets and computes a new vector.

    Args:
        svd_dict (dict): A dictionary containing SVD components for multiple datasets. The structure of the dictionary is expected to be:
                         {
                             dataset_name: {
                                 key: {
                                     "u": tensor,
                                     "s": tensor,
                                     "v": tensor,
                                     "dim1": tensor (optional)
                                 }
                             }
                         }
        config (object): A configuration object that contains a list of dataset names under the attribute `DATASETS`.

    Returns:
        dict: A dictionary containing the merged SVD components or averaged "dim1" values for each key.
    """
    print("Summing SVD...")
    new_vector = {}
    for key in svd_dict[config.DATASETS[0]]:
        if "u" in svd_dict[config.DATASETS[0]][key].keys():
            sum_u = sum([svd_dict[dataset][key]["u"] for dataset in config.DATASETS])
            sum_s = sum([svd_dict[dataset][key]["s"] for dataset in config.DATASETS])
            sum_v = sum([svd_dict[dataset][key]["v"] for dataset in config.DATASETS])

            u_u, s_u, v_u = torch.linalg.svd(sum_u, full_matrices=False)
            u_v, s_v, v_v = torch.linalg.svd(sum_v, full_matrices=False)
            new_vector[key] = torch.linalg.multi_dot(
                (
                    u_u,
                    v_u,
                    torch.diag(sum_s),
                    u_v,
                    v_v,
                )
            )
        else:
            for i, dataset in enumerate(config.DATASETS, start=1):
                if i == 1:
                    new_vector[key] = svd_dict[dataset][key]["dim1"]
                else:
                    new_vector[key] += (
                        svd_dict[dataset][key]["dim1"] - new_vector[key]
                    ) / i
    return new_vector


###############
##### LOSSLESS Orthogonalization
def compute_and_sum_svd_mem_reduction_lossless(
    task_vectors: List[StateDictType],
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
):
    """
    Computes the Singular Value Decomposition (SVD) for each task vector and merge the results.

    This function performs the following steps:
    1. Iterates over each layer in the task vectors.
    2. For each layer, it computes the SVD of the corresponding matrix if it is a 2D tensor excluding "text_projection".
    3. Concatenate the U_i, S_i, and V_i matrices from the SVD across all tasks.
    4. If the vector is not a 2D tensor or is "text_projection", it computes the mean of the vectors.
    5. After concatenating the SVD components, recomputes the SVD of the summed U and V matrices and constructs the merged layer.

    Args:
        task_vectors (list): A list of task vectors, where each task vector is a dictionary containing the vectors for each task.
        accelerator (torch.device): The device to use for the computation.
    Returns:
        dict: A dictionary containing the new vectors after summing the SVD components.
    """
    # becareful wit vit-l on 20 task it does not fit in GPU or in 64 GB RAM (try without last layer)
    print("Computing SVD...")
    with torch.no_grad():
        new_vector = {}
        for key in task_vectors[0]:
            original_device = task_vectors[0][key].device
            new_vector[key] = {}
            for i, task_vector in enumerate(task_vectors):
                vec = task_vector[key].to(accelerator)

                if len(task_vector[key].shape) == 2 and "text_projection" not in key:

                    u, s, v = torch.linalg.svd(vec, full_matrices=False)

                    if i == 0:
                        print(f"Computed SVD for {key}...")
                        sum_u = torch.zeros(
                            u.shape[0],
                            u.shape[1] * len(task_vectors),
                            device=accelerator,
                        )
                        sum_s = torch.zeros(
                            s.shape[0] * len(task_vectors), device=accelerator
                        )
                        sum_v = torch.zeros(
                            v.shape[0] * len(task_vectors),
                            v.shape[1],
                            device=accelerator,
                        )
                    reduced_index_s = s.shape[0]

                    # select only the first reduced_index_s columns of u and place them
                    sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                        :, :reduced_index_s
                    ]
                    sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                        :reduced_index_s
                    ]
                    # select only the first reduced_index_s rows of v and place them
                    sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                        :reduced_index_s, :
                    ]

                else:
                    if i == 0:
                        new_vector[key] = vec.clone()
                    else:
                        new_vector[key] += (vec - new_vector[key]) / (i + 1)

            if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                u_u, s_u, v_u = torch.linalg.svd(sum_u, full_matrices=False)
                u_v, s_v, v_v = torch.linalg.svd(sum_v, full_matrices=False)

                new_vector[key] = torch.linalg.multi_dot(
                    (
                        u_u,
                        v_u,
                        torch.diag(sum_s),
                        u_v,
                        v_v,
                    )
                )
            new_vector[key] = new_vector[key].to(original_device, non_blocking=True)
    return new_vector


###############
##### LOSSLESS EIGENDECOMP
def compute_and_sum_svd_mem_reduction_lossless_eigen(
    task_vectors: List[StateDictType],
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
):
    """
    Computes the Singular Value Decomposition (SVD) for each task vector and merge the results.

    This function performs the following steps:
    1. Iterates over each layer in the task vectors.
    2. For each layer, it computes the SVD of the corresponding matrix if it is a 2D tensor excluding "text_projection".
    3. Concatenate the U_i, S_i, and V_i matrices from the SVD across all tasks.
    4. If the vector is not a 2D tensor or is "text_projection", it computes the mean of the vectors.
    5. After concatenating the SVD components, recomputes the eigendecomposition of the summed U and V matrices and constructs the merged layer.

    Args:
        task_vectors (list): A list of task vectors, where each task vector is a dictionary containing the vectors for each task.
        accelerator (torch.device): The device to use for the computation.

    Returns:
        dict: A dictionary containing the new vectors after merging the SVD components.
    """
    # becareful wit vit-l on 20 task it does not fit in GPU or in 64 GB RAM (try without last layer)
    print("Computing SVD...")
    with torch.no_grad():
        new_vector = {}
        for key in task_vectors[0]:
            original_device = task_vectors[0][key].device
            new_vector[key] = {}
            for i, task_vector in enumerate(task_vectors):
                vec = task_vector[key].to(accelerator)

                if len(task_vector[key].shape) == 2 and "text_projection" not in key:

                    u, s, v = torch.linalg.svd(vec, full_matrices=False)

                    if i == 0:
                        print(f"Computed SVD for {key}...")
                        sum_u = torch.zeros(
                            u.shape[0],
                            u.shape[1] * len(task_vectors),
                            device=accelerator,
                        )
                        sum_s = torch.zeros(
                            s.shape[0] * len(task_vectors), device=accelerator
                        )
                        sum_v = torch.zeros(
                            v.shape[0] * len(task_vectors),
                            v.shape[1],
                            device=accelerator,
                        )
                    reduced_index_s = s.shape[0]

                    # select only the first reduced_index_s columns of u and place them
                    sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                        :, :reduced_index_s
                    ]
                    sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                        :reduced_index_s
                    ]
                    # select only the first reduced_index_s rows of v and place them
                    sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                        :reduced_index_s, :
                    ]

                else:
                    if i == 0:
                        new_vector[key] = vec.clone()
                    else:
                        new_vector[key] += (vec - new_vector[key]) / (i + 1)

            if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                sum_s, indices = torch.sort(sum_s, stable=True)

                sum_u = torch.index_select(sum_u, 1, indices)
                l_u, q_u = torch.linalg.eigh(sum_u.mT @ sum_u)
                u_orth = (
                    q_u
                    @ torch.diag(1.0 / (torch.sqrt(torch.abs(l_u)) + 1e-12))
                    @ q_u.mT
                )

                sum_v = torch.index_select(sum_v, 0, indices)

                l_v, q_v = torch.linalg.eigh(sum_v @ sum_v.mT)
                v_orth = (
                    q_v
                    @ torch.diag(1.0 / (torch.sqrt(torch.abs(l_v)) + 1e-12))
                    @ q_v.mT
                )

                new_vector[key] = torch.linalg.multi_dot(  # bool_mask *
                    (
                        sum_u,
                        u_orth,
                        torch.diag(sum_s),
                        v_orth,
                        sum_v,
                    )
                )
            new_vector[key] = new_vector[key].to(original_device, non_blocking=True)
    return new_vector


###############
#### TSV Merge Orthogonalization
@torch.no_grad()
def compute_and_sum_svd_mem_reduction(
    task_vectors: List[StateDictType],
    exclude_keys: Optional[List[str]] = None,
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
    return_single_task_models: bool = False,
):
    """
    Computes the Singular Value Decomposition (SVD) for each vector in the task_vectors,
    reduces the dimensionality of the vectors based on the sv_reduction factor, and concatenate
    the low-rank matrices. If the vector is not a 2D tensor or is "text_projection", it computes the mean of the vectors.
    Computation of the SVD is performed also for the second operation.

    Args:
        task_vectors (list): A list of task vector objects, where each object contains a
                            dictionary of vectors.
        exclude_keys (list): A list of keys to exclude from the TSVM.
        accelerator (torch.device): The device to use for the computation.
        return_single_task_models (bool): Whether to return the single task models after the TSVM.

    Returns:
        dict: A dictionary containing the new vectors after SVD computation and merging.
    """
    if exclude_keys is None:
        exclude_keys = []
    num_tasks = len(task_vectors)
    sv_reduction = 1 / num_tasks

    new_vector: Dict[str, Tensor] = {}
    if return_single_task_models:
        single_task_models = [{} for _ in range(num_tasks)]
    for key in task_vectors[0]:
        original_device = task_vectors[0][key].device
        original_dtype = task_vectors[0][key].dtype

        for i, task_vector in enumerate(task_vectors):
            vec = task_vector[key].to(device=accelerator, non_blocking=True)
            if len(task_vector[key].shape) == 2 and key not in exclude_keys:
                # at current, the SVD is not supported for half precision, so we need to convert to float32
                if not (
                    original_dtype == torch.float32 or original_dtype == torch.float64
                ):
                    vec = vec.to(dtype=torch.float32)

                # vec = u @ torch.diag(s) @ vh
                u, s, vh = torch.linalg.svd(vec, full_matrices=False)

                if i == 0:
                    print(f"Computed SVD for {key}...")
                    sum_u = torch.zeros_like(u, device=accelerator)
                    sum_s = torch.zeros_like(s, device=accelerator)
                    sum_vh = torch.zeros_like(vh, device=accelerator)
                reduced_index_s = int(s.shape[0] * sv_reduction)

                # select only the first reduced_index_s columns of u and place them
                sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                    :, :reduced_index_s
                ]
                sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                    :reduced_index_s
                ]
                # select only the first reduced_index_s rows of v and place them
                sum_vh[i * reduced_index_s : (i + 1) * reduced_index_s, :] = vh[
                    :reduced_index_s, :
                ]
            else:
                # if the vector is not a 2D tensor or is in exclude_keys, compute the mean
                if i == 0:
                    new_vector[key] = vec.clone()
                else:
                    new_vector[key] += (vec - new_vector[key]) / (i + 1)

        if len(task_vector[key].shape) == 2 and key not in exclude_keys:
            u_u, s_u, vh_u = torch.linalg.svd(sum_u, full_matrices=False)
            u_vh, s_vh, vh_vh = torch.linalg.svd(sum_vh, full_matrices=False)

            new_vector[key] = torch.linalg.multi_dot(
                (
                    u_u,
                    vh_u,
                    torch.diag(sum_s),
                    u_vh,
                    vh_vh,
                )
            )
            new_vector[key] = new_vector[key].to(
                device=original_device, dtype=original_dtype, non_blocking=True
            )
            if return_single_task_models:
                reduced_index_s = int(sum_s.shape[0] * sv_reduction)
                new_u = u_u @ vh_u
                new_vh = u_vh @ vh_vh
                for i in range(num_tasks):
                    single_task_models[i][key] = torch.linalg.multi_dot(
                        (
                            new_u[:, i * reduced_index_s : (i + 1) * reduced_index_s],
                            torch.diag(
                                sum_s[i * reduced_index_s : (i + 1) * reduced_index_s]
                            ),
                            new_vh[i * reduced_index_s : (i + 1) * reduced_index_s, :],
                        )
                    ).to(
                        device=original_device, dtype=original_dtype, non_blocking=True
                    )
        else:
            new_vector[key] = new_vector[key].to(
                device=original_device, dtype=original_dtype, non_blocking=True
            )
            if return_single_task_models:
                for i in range(num_tasks):
                    single_task_models[i][key] = new_vector[key].clone()

    if not return_single_task_models:
        return new_vector
    else:
        return new_vector, single_task_models


###############
#### TSV Merge Eigendecomp
def compute_and_sum_svd_mem_reduction_2(
    task_vectors: List[StateDictType],
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
):
    """
    Computes the Singular Value Decomposition (SVD) for each vector in the task_vectors,
    reduces the dimensionality of the vectors based on the sv_reduction factor, and concatenate
    the low-rank matrices. If the vector is not a 2D tensor or is "text_projection", it computes the mean of the vectors.
    Computation of the eigendecomposition is performed instead of the SVD for the second operation.

    Args:
        task_vectors (list): A list of task vector objects, where each object contains a
                             dictionary of vectors.
        accelerator (torch.device): The device to use for the computation.

    Returns:
        dict: A dictionary containing the new vectors after SVD computation and merging.
    """
    sv_reduction = 1 / len(task_vectors)

    print("Computing SVD...")
    with torch.no_grad():
        new_vector = {}
        for key in task_vectors[0]:
            original_device = task_vectors[0][key].device
            new_vector[key] = {}
            for i, task_vector in enumerate(task_vectors):
                vec = task_vector[key].to(accelerator)

                if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                    u, s, v = torch.linalg.svd(vec, full_matrices=False)

                    if i == 0:
                        print(f"Computed SVD for {key}...")
                        sum_u = torch.zeros_like(u, device=accelerator)
                        sum_s = torch.zeros_like(s, device=accelerator)
                        sum_v = torch.zeros_like(v, device=accelerator)
                    reduced_index_s = int(s.shape[0] * sv_reduction)

                    # select only the first reduced_index_s columns of u and place them
                    sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                        :, :reduced_index_s
                    ]
                    sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                        :reduced_index_s
                    ]
                    # select only the first reduced_index_s rows of v and place them
                    sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                        :reduced_index_s, :
                    ]

                else:
                    if i == 0:
                        new_vector[key] = vec.clone()
                    else:
                        new_vector[key] += (vec - new_vector[key]) / (i + 1)

            if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                sum_s, indices = torch.sort(sum_s, stable=True)

                sum_u = torch.index_select(sum_u, 1, indices)
                l_u, q_u = torch.linalg.eigh(sum_u.mT @ sum_u)
                u_orth = (
                    q_u
                    @ torch.diag(1.0 / (torch.sqrt(torch.abs(l_u)) + 1e-12))
                    @ q_u.mT
                )

                sum_v = torch.index_select(sum_v, 0, indices)

                l_v, q_v = torch.linalg.eigh(sum_v @ sum_v.mT)
                v_orth = (
                    q_v
                    @ torch.diag(1.0 / (torch.sqrt(torch.abs(l_v)) + 1e-12))
                    @ q_v.mT
                )

                new_vector[key] = torch.linalg.multi_dot(  # bool_mask *
                    (
                        sum_u,
                        u_orth,
                        torch.diag(sum_s),
                        v_orth,
                        sum_v,
                    )
                )
            new_vector[key] = new_vector[key].to(original_device, non_blocking=True)

    return new_vector


###############
#### Rank Reduction TV
def compute_and_sum_svd_mem_reduction_rank_reduction(
    task_vectors: List[StateDictType],
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
):
    """
    Compute and sum the Singular Value Decomposition (SVD) of task vectors with rank reduction.

    This function performs SVD on the vectors in `task_vectors` and reduces their rank based on the
    number of tasks specified in `config.DATASETS`. The reduced vectors are then summed together.

    Args:
        task_vectors (list): A list of task vector objects. Each object should have a `vector` attribute
                             which is a dictionary where keys are vector names and values are tensors.
        accelerator (torch.device): The device to use for the computation.

    Returns:
        dict: A dictionary containing the new vectors after SVD computation and summation.
    """
    sv_reduction = 1 / len(task_vectors)
    print("Computing SVD...")
    with torch.no_grad():
        new_vector = {}
        for key in task_vectors[0]:
            original_device = task_vectors[0][key].device
            new_vector[key] = {}
            for i, task_vector in enumerate(task_vectors):
                vec = task_vector[key].to(accelerator)

                if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                    u, s, v = torch.linalg.svd(vec, full_matrices=False)

                    if i == 0:
                        print(f"Computed SVD for {key}...")
                        sum_u = torch.zeros_like(u, device=accelerator)
                        sum_s = torch.zeros_like(s, device=accelerator)
                        sum_v = torch.zeros_like(v, device=accelerator)
                    reduced_index_s = int(s.shape[0] * sv_reduction)

                    # select only the first reduced_index_s columns of u and place them
                    sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                        :, :reduced_index_s
                    ]
                    sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                        :reduced_index_s
                    ]
                    # select only the first reduced_index_s rows of v and place them
                    sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                        :reduced_index_s, :
                    ]

                else:
                    if i == 0:
                        new_vector[key] = vec.clone()
                    else:
                        new_vector[key] += (vec - new_vector[key]) / (i + 1)

            if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                new_vector[key] = torch.linalg.multi_dot(
                    (
                        sum_u,
                        torch.diag(sum_s),
                        sum_v,
                    )
                )

            new_vector[key] = new_vector[key].to(original_device, non_blocking=True)
    return new_vector


def compute_and_sum_svd_mem_reduction_dummy(
    task_vectors: List[StateDictType],
    accelerator: torch.device = "cuda" if torch.cuda.is_available() else "cpu",
):
    """To perform dummy operations."""
    sv_reduction = 1 / len(task_vectors)
    print("Computing SVD...")
    with torch.no_grad():
        new_vector = {}
        for key in task_vectors[0]:
            original_device = task_vectors[0][key].device
            new_vector[key] = {}
            for i, task_vector in enumerate(task_vectors):
                vec = task_vector[key].to(accelerator)

                if len(task_vector[key].shape) == 2 and "text_projection" not in key:
                    if i == 0:
                        u, s, v = torch.linalg.svd(vec, full_matrices=False)
                        reduced_index_s = int(s.shape[0] * sv_reduction)

                        print(f"Computed SVD for {key}...")
                        sum_u = torch.zeros_like(u)
                        sum_s = torch.zeros_like(s)
                        sum_v = torch.zeros_like(v)

                        # select only the first reduced_index_s columns of u and place them
                        sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                            :, :reduced_index_s
                        ]
                        sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                            :reduced_index_s
                        ]
                        # select only the first reduced_index_s rows of v and place them
                        sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                            :reduced_index_s, :
                        ]
                    else:
                        # generate u vectors orthogonal to the previous ones
                        # generate v vectors orthogonal to the previous ones
                        print("dummy")
                        u = torch.nn.functional.normalize(
                            torch.randn_like(sum_u), p=2, dim=-2
                        )
                        v = torch.nn.functional.normalize(
                            torch.randn_like(sum_v), p=2, dim=-1
                        )

                        # select only the first reduced_index_s columns of u and place them
                        sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                            :, :reduced_index_s
                        ]
                        sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                            :reduced_index_s
                        ]
                        # select only the first reduced_index_s rows of v and place them
                        sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                            :reduced_index_s, :
                        ]

                else:
                    if i == 0:
                        new_vector[key] = vec.clone()
                    else:
                        new_vector[key] += (vec - new_vector[key]) / (i + 1)

            if len(task_vector[key].shape) == 2 and "text_projection" not in key:

                new_vector[key] = torch.linalg.multi_dot(
                    (
                        sum_u,
                        torch.diag(sum_s),
                        sum_v,
                    )
                )

            new_vector[key] = new_vector[key].to(original_device, non_blocking=True)
    return new_vector



classnames = [
    "angry",
    "disgusted",
    "fearful",
    "happy",
    "neutral",
    "sad",
    "surprised",
]

templates = [
    lambda c: f"a photo of a {c} looking face.",
    lambda c: f"a photo of a face showing the emotion: {c}.",
    lambda c: f"a photo of a face looking {c}.",
    lambda c: f"a face that looks {c}.",
    lambda c: f"they look {c}.",
    lambda c: f"look at how {c} they are.",
]



from collections import OrderedDict
from numbers import Number
from typing import Callable, Dict, List, Literal, Union, cast

import torch
from torch import Tensor
from tqdm.auto import tqdm

from .parameters import check_parameters_all_equal
from .type import BoolStateDictType, StateDictType


def to_device(
    state_dict: StateDictType,
    device: Union[torch.device, str],
    copy: bool = False,
    inplace: bool = False,
):
    if inplace:
        ret_state_dict = state_dict
    else:
        ret_state_dict = OrderedDict()
    for key in state_dict:
        ret_state_dict[key] = cast(Tensor, state_dict[key]).to(
            device, non_blocking=True, copy=copy
        )
    return ret_state_dict


def state_dicts_check_keys(state_dicts: List[StateDictType]):
    """
    Checks that the state dictionaries have the same keys.

    Args:
        state_dicts (List[Dict[str, Tensor]]): A list of dictionaries containing the state of PyTorch models.

    Raises:
        ValueError: If the state dictionaries have different keys.
    """
    # Get the keys of the first state dictionary in the list
    keys = set(state_dicts[0].keys())
    # Check that all the state dictionaries have the same keys
    for state_dict in state_dicts:
        assert keys == set(state_dict.keys()), "keys of state_dicts are not equal"


def num_params_of_state_dict(state_dict: StateDictType):
    """
    Returns the number of parameters in a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to count the number of parameters in.

    Returns:
        int: The number of parameters in the state dict.
    """
    return sum([state_dict[key].numel() for key in state_dict])


def state_dict_flatten(state_dict: Dict[str, Tensor]):
    """
    Flattens a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to be flattened.

    Returns:
        Tensor: The flattened state dict.
    """
    flattened_state_dict = []
    for key in state_dict:
        flattened_state_dict.append(state_dict[key].flatten())
    return torch.cat(flattened_state_dict)


def state_dict_avg(state_dicts: List[StateDictType]):
    """
    Returns the average of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to average.

    Returns:
        Dict: The average of the state dicts.
    """
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    num_state_dicts = len(state_dicts)
    avg_state_dict = OrderedDict()
    for key in state_dicts[0]:
        avg_state_dict[key] = torch.zeros_like(state_dicts[0][key])
        for state_dict in state_dicts:
            avg_state_dict[key] += state_dict[key]
        avg_state_dict[key] /= num_state_dicts
    return avg_state_dict


def state_dict_sub(
    a: StateDictType, b: StateDictType, strict: bool = True, device=None
):
    """
    Returns the difference between two state dicts `a-b`.

    Args:
        a (StateDictType): The first state dict.
        b (StateDictType): The second state dict.
        strict (bool): Whether to check if the keys of the two state dicts are the same.

    Returns:
        StateDictType: The difference between the two state dicts.
    """
    if strict:
        assert set(a.keys()) == set(b.keys())

    diff = OrderedDict()
    for k in a:
        if k in b:
            diff[k] = a[k] - b[k]
            if device is not None:
                diff[k] = diff[k].to(device, non_blocking=True)
    return diff


def state_dict_add(
    a: StateDictType,
    b: StateDictType,
    strict: bool = True,
    device=None,
    show_pbar: bool = False,
):
    """
    Returns the sum of two state dicts.

    Args:
        a (Dict): The first state dict.
        b (Dict): The second state dict.
        strict (bool): Whether to check if the keys of the two state dicts are the same.

    Returns:
        Dict: The sum of the two state dicts.
    """
    ans = {}
    if strict:
        check_parameters_all_equal([a, b])
        for key in tqdm(tuple(a.keys())) if show_pbar else a:
            ans[key] = a[key] + b[key]
    else:
        for key in tqdm(tuple(a.keys())) if show_pbar else a:
            if key in b:
                ans[key] = a[key] + b[key]
    if device is not None:
        ans = to_device(ans, device)
    return ans


def state_dict_add_scalar(a: StateDictType, scalar: Number):
    ans = OrderedDict()
    for key in a:
        ans[key] = a[key] + scalar
    return ans


def state_dict_mul(state_dict: StateDictType, scalar: float):
    """
    Returns the product of a state dict and a scalar.

    Args:
        state_dict (Dict): The state dict to be multiplied.
        scalar (float): The scalar to multiply the state dict with.

    Returns:
        Dict: The product of the state dict and the scalar.
    """
    diff = OrderedDict()
    for k in state_dict:
        diff[k] = scalar * state_dict[k]
    return diff


def state_dict_div(state_dict: StateDictType, scalar: float, show_pbar: bool = False):
    """
    Returns the division of a state dict by a scalar.

    Args:
        state_dict (Dict): The state dict to be divided.
        scalar (float): The scalar to divide the state dict by.

    Returns:
        Dict: The division of the state dict by the scalar.
    """
    diff = OrderedDict()
    for k in tqdm(tuple(state_dict.keys())) if show_pbar else state_dict:
        diff[k] = state_dict[k] / scalar
    return diff


def state_dict_power(state_dict: Dict[str, Tensor], p: float):
    """
    Returns the power of a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to be powered.
        p (float): The power to raise the state dict to.

    Returns:
        Dict[str, Tensor]: The powered state dict.
    """
    powered_state_dict = {}
    for key in state_dict:
        powered_state_dict[key] = state_dict[key] ** p
    return powered_state_dict


def state_dict_interpolation(
    state_dicts: List[Dict[str, Tensor]], scalars: List[float]
):
    """
    Interpolates between a list of state dicts using a list of scalars.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to interpolate between.
        scalars (List[float]): The list of scalars to use for interpolation.

    Returns:
        Dict: The interpolated state dict.
    """
    assert len(state_dicts) == len(
        scalars
    ), "The number of state_dicts and scalars must be the same"
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    interpolated_state_dict = {}
    for key in state_dicts[0]:
        interpolated_state_dict[key] = torch.zeros_like(state_dicts[0][key])
        for state_dict, scalar in zip(state_dicts, scalars):
            interpolated_state_dict[key] += scalar * state_dict[key]
    return interpolated_state_dict


def state_dict_sum(state_dicts: List[StateDictType]):
    """
    Returns the sum of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to sum.

    Returns:
        Dict: The sum of the state dicts.
    """
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    sum_state_dict = OrderedDict()
    for key in state_dicts[0]:
        sum_state_dict[key] = 0
        for state_dict in state_dicts:
            sum_state_dict[key] = sum_state_dict[key] + state_dict[key]
    return sum_state_dict


def state_dict_weighted_sum(
    state_dicts: List[Dict[str, Tensor]], weights: List[float], device=None
):
    """
    Returns the weighted sum of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to interpolate between.
        weights (List[float]): The list of weights to use for the weighted sum.

    Returns:
        Dict: The weighted sum of the state dicts.
    """
    assert len(state_dicts) == len(
        weights
    ), "The number of state_dicts and weights must be the same"
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    weighted_sum_state_dict: Dict[str, Tensor] = {}
    for key in state_dicts[0]:
        # states dicts can be sparse matrices
        weighted_sum_state_dict[key] = torch.zeros_like(state_dicts[0][key]).to_dense()
        for state_dict, weight in zip(state_dicts, weights):
            weighted_sum_state_dict[key] = torch.add(
                weighted_sum_state_dict[key], weight * state_dict[key]
            )
        if device is not None:
            weighted_sum_state_dict[key] = weighted_sum_state_dict[key].to(
                device, non_blocking=True
            )
    return weighted_sum_state_dict


def state_dict_diff_abs(a: StateDictType, b: StateDictType):
    """
    Returns the per-layer abs of the difference between two state dicts.

    Args:
        a (StateDictType): The first state dict.
        b (StateDictType): The second state dict.

    Returns:
        StateDictType: The absolute difference between the two state dicts.
    """
    diff = state_dict_sub(a, b)
    abs_diff = {key: diff[key].abs() for key in diff}
    return abs_diff


def state_dict_binary_mask(
    a: StateDictType,
    b: StateDictType,
    compare_fn: Union[
        Literal["greater", "less", "equal", "not_equal"],
        Callable[[Tensor, Tensor], torch.BoolTensor],
    ] = "greater",
) -> BoolStateDictType:
    """
    Returns the binary mask of elements in a compared to elements in b using the provided comparison function.

    Args:
        a (StateDictType): The first state dict.
        b (StateDictType): The second state dict.
        compare_fn (Union[Literal["greater", "less", "equal", "not_equal"], Callable[[Tensor, Tensor], Tensor]]): A function that takes two tensors and returns a boolean tensor.
            Defaults to greater than comparison (x > y).

    Returns:
        StateDictType: A dictionary containing binary masks (0 or 1) based on the comparison.
    """
    compare_fn_dict = {
        "greater": lambda x, y: x > y,
        "less": lambda x, y: x < y,
        "equal": lambda x, y: x == y,
        "not_equal": lambda x, y: x != y,
    }
    if isinstance(compare_fn, str):
        compare_fn = compare_fn_dict[compare_fn]
    elif not callable(compare_fn):
        raise ValueError(
            f"compare_fn must be a string or a callable, but got {type(compare_fn)}"
        )

    mask = OrderedDict()
    for key in a:
        mask[key] = compare_fn(a[key], b[key])
    return mask


def state_dict_hadmard_product(a: StateDictType, b: StateDictType) -> StateDictType:
    """
    Returns the Hadamard product of two state dicts, i.e. element-wise product.

    Args:
        a (StateDictType): The first state dict.
        b (StateDictType): The second state dict.

    Returns:
        StateDictType: The Hadamard product of the two state dicts.
    """
    ans = OrderedDict()
    for key in a:
        ans[key] = a[key] * b[key]
    return ans



import os
from typing import cast

import torch
from huggingface_hub import hf_hub_download
from huggingface_hub.utils import EntryNotFoundError
from torch import Tensor, nn
from transformers import CLIPModel, CLIPProcessor
from trl.import_utils import is_npu_available, is_xpu_available


class MLP(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(768, 1024),
            nn.Dropout(0.2),
            nn.Linear(1024, 128),
            nn.Dropout(0.2),
            nn.Linear(128, 64),
            nn.Dropout(0.1),
            nn.Linear(64, 16),
            nn.Linear(16, 1),
        )

    @torch.no_grad()
    def forward(self, embed: Tensor) -> Tensor:
        """
        Forward pass through the MLP. The return value is a single scalar.
        """
        return self.layers(embed)


class AestheticScorer(torch.nn.Module):
    """
    This model attempts to predict the aesthetic score of an image. The aesthetic score
    is a numerical approximation of how much a specific image is liked by humans on average.
    This is from https://github.com/christophschuhmann/improved-aesthetic-predictor

    Note for `model_id` and `model_filename`:

        In some implementation, the filename of the MLP model is 'sac+logos+ava1-l14-linearMSE.pth',
        which is the same as the default value of the 'model_filename' parameter in the constructor ('aesthetic-model.pth').
        It was simply renamed to 'aesthetic-model.pth' in the implementation.
        see https://huggingface.co/trl-lib/ddpo-aesthetic-predictor/commit/7f639699bec8126062148a47ecb1a4312d8e6688
    """

    def __init__(
        self,
        *,
        dtype: torch.dtype,
        model_id: str = "trl-lib/ddpo-aesthetic-predictor",
        model_filename: str = "aesthetic-model.pth",
    ):
        """
        Initialize the AestheticScorer class.

        Args:
            dtype (torch.dtype): The data type of the tensors.
            model_id (str, optional): The ID of the model to download. Defaults to "trl-lib/ddpo-aesthetic-predictor".
            model_filename (str, optional): The filename of the model to download. Defaults to "aesthetic-model.pth". This is the same as 'sac+logos+ava1-l14-linearMSE.pth' in some implementations.
        """
        super().__init__()
        self.clip = CLIPModel.from_pretrained("openai/clip-vit-large-patch14")
        self.processor = CLIPProcessor.from_pretrained("openai/clip-vit-large-patch14")
        self.mlp = MLP()
        try:
            cached_path = hf_hub_download(model_id, model_filename)
        except EntryNotFoundError:
            cached_path = os.path.join(model_id, model_filename)
        state_dict = torch.load(cached_path, map_location=torch.device("cpu"))
        self.mlp.load_state_dict(state_dict)
        self.dtype = dtype
        self.eval()

    @torch.no_grad()
    def __call__(self, images: Tensor) -> Tensor:
        """
        Process the given images and return their aesthetic scores.

        This method processes the images using the CLIP model, normalizes the embeddings,
        and then passes them through a MLP to get the aesthetic scores.

        Args:
            images (torch.Tensor): A batch of images to process.

        Returns:
            Tensor: The aesthetic scores of the images. Return shape is (batch_size,).
        """
        device = next(self.parameters()).device
        inputs = self.processor(images=images, return_tensors="pt")
        inputs = {
            k: cast(Tensor, v).to(self.dtype).to(device) for k, v in inputs.items()
        }
        embed = self.clip.get_image_features(**inputs)
        # normalize embedding
        embed = embed / torch.linalg.vector_norm(embed, dim=-1, keepdim=True)
        return self.mlp(embed).squeeze(1)


def aesthetic_scorer(
    dtype: torch.dtype = torch.float32,
    hub_model_id: str = "trl-lib/ddpo-aesthetic-predictor",
    model_filename: str = "aesthetic-model.pth",
):
    scorer = AestheticScorer(
        dtype=dtype,
        model_id=hub_model_id,
        model_filename=model_filename,
    )
    if is_npu_available():
        scorer = scorer.npu()
    elif is_xpu_available():
        scorer = scorer.xpu()
    else:
        scorer = scorer.cuda()

    def _fn(images: Tensor, prompts, metadata):
        images = (images * 255).round().clamp(0, 255).to(torch.uint8)
        scores: Tensor = scorer(images)
        return scores, {}

    return _fn



import json
import logging
import os
from copy import deepcopy
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple, Type

import torch
from accelerate import init_empty_weights
from accelerate.utils.constants import SAFE_WEIGHTS_NAME, WEIGHTS_NAME
from huggingface_hub import snapshot_download
from safetensors import safe_open
from safetensors.torch import load_file
from torch import nn
from transformers import AutoConfig

from fusion_bench.utils.dtype import parse_dtype
from fusion_bench.utils.packages import import_object

if TYPE_CHECKING:
    from transformers import PretrainedConfig

log = logging.getLogger(__name__)

__all__ = ["resolve_checkpoint_path", "LazyStateDict"]


def resolve_checkpoint_path(
    checkpoint: str,
    hf_revision: Optional[str] = None,
    hf_cache_dir: Optional[str] = None,
    hf_proxies: Optional[Dict] = None,
):
    # If it's a local file or directory, return as is
    if os.path.exists(checkpoint):
        return checkpoint
    # If it's a HuggingFace Hub model id, download snapshot
    try:
        # This will download the model to the cache and return the local path
        local_path = snapshot_download(
            repo_id=checkpoint,
            revision=hf_revision,
            cache_dir=hf_cache_dir,
            proxies=hf_proxies,
        )
        return local_path
    except Exception as e:
        raise FileNotFoundError(
            f"Could not resolve checkpoint: {checkpoint}. Error: {e}"
        )


class LazyStateDict:
    """
    Dictionary-like object that lazily loads a state dict from a checkpoint path.
    """

    _local_path: str
    """local path to the checkpoint."""
    _state_dict_cache: Optional[Dict]
    """Cache for the state dict, if enabled."""
    _index_filename: Optional[str]
    _checkpoint_files: Optional[List[str]]
    _index: Optional[Dict[str, str]]
    """Mapping of parameter names to checkpoint files."""

    def __init__(
        self,
        checkpoint: str,
        meta_module_class: Optional[Type[nn.Module]] = None,
        meta_module: Optional[nn.Module] = None,
        cache_state_dict: bool = False,
        torch_dtype: Optional[torch.dtype] = None,
        device: str = "cpu",
        hf_revision: Optional[str] = None,
        hf_cache_dir: Optional[str] = None,
        hf_proxies: Optional[Dict] = None,
    ):
        """
        Args:
            checkpoint (str): Path to the checkpoint file or directory.
            meta_module_class (Type[nn.Module], optional): Class of the meta module to instantiate.
            meta_module (nn.Module, optional): Pre-initialized meta module.
            cache_state_dict (bool): Whether to cache the state dict in memory.
            torch_dtype (torch.dtype, optional): The dtype to use for the tensors.
            device (str): The device to load the tensors onto.
            hf_revision (str, optional): The revision of the model to download from Hugging Face Hub.
            hf_cache_dir (str, optional): The cache directory for Hugging Face models.
            hf_proxies (Dict, optional): Proxies to use for downloading from Hugging Face Hub.
        """
        self.cache_state_dict = cache_state_dict
        self.meta_module_class = meta_module_class
        if isinstance(self.meta_module_class, str):
            self.meta_module_class = import_object(self.meta_module_class)
        self.meta_module = meta_module
        if self.meta_module_class is not None:
            if self.meta_module is not None:
                raise ValueError(
                    "Cannot provide both meta_module_class and meta_module, please provide only one."
                )
            with init_empty_weights():
                self.meta_module = self.meta_module_class.from_pretrained(
                    checkpoint,
                    torch_dtype=torch_dtype,
                    revision=hf_revision,
                    cache_dir=hf_cache_dir,
                    proxies=hf_proxies,
                )

        self._checkpoint = checkpoint
        self._local_path = resolve_checkpoint_path(
            checkpoint,
            hf_revision=hf_revision,
            hf_cache_dir=hf_cache_dir,
            hf_proxies=hf_proxies,
        )

        self._index, self._index_filename, self._checkpoint_files = (
            self._resolve_checkpoint_files(self._local_path)
        )

        if self._index is not None:
            # if meta_module is provided, remove the keys that are not in the meta_module
            if self.meta_module is not None:
                meta_module_state_dict = self.meta_module.state_dict()
                for key in tuple(self._index.keys()):
                    if key not in meta_module_state_dict:
                        self._index.pop(key)
            if cache_state_dict:
                self._state_dict_cache = {}
            else:
                self._state_dict_cache = None
        elif len(self._checkpoint_files) == 1 and self._checkpoint_files[0].endswith(
            SAFE_WEIGHTS_NAME
        ):
            # let the keys of self._index be the keys of the state dict, the values are the checkpoint file
            with safe_open(
                self._checkpoint_files[0], framework="pt", device=device
            ) as f:
                self._index = {key: self._checkpoint_files[0] for key in f.keys()}
                if cache_state_dict:
                    self._state_dict_cache = {}
                else:
                    self._state_dict_cache = None
        elif len(self._checkpoint_files) == 1 and self._checkpoint_files[0].endswith(
            WEIGHTS_NAME
        ):
            log.info(f"Loading full state dict from {WEIGHTS_NAME}")
            self._state_dict_cache = torch.load(self._checkpoint_files[0])
            # if meta_module is provided, remove the keys that are not in the meta_module
            if self.meta_module is not None:
                meta_module_state_dict = self.meta_module.state_dict()
                for key in tuple(self._state_dict_cache.keys()):
                    if key not in meta_module_state_dict:
                        self._state_dict_cache.pop(key)
        else:
            raise ValueError(
                f"Cannot determine the type of checkpoint, please provide a checkpoint path to a file containing a whole state dict with file name {WEIGHTS_NAME} or {SAFE_WEIGHTS_NAME}, or the index of a sharded checkpoint ending with `.index.json`."
            )

        self._torch_dtype = parse_dtype(torch_dtype)
        self._device = device

    @property
    def checkpoint(self) -> str:
        return self._checkpoint

    @property
    def config(self) -> "PretrainedConfig":
        return AutoConfig.from_pretrained(self._checkpoint)

    def state_dict(self, keep_vars: bool = False) -> "LazyStateDict":
        """
        Args:
            keep_vars (bool): Ignored, as LazyStateDict does not support keep_vars. Just for compatibility.
        """
        return self

    def _resolve_checkpoint_files(self, checkpoint: str):
        # reference: https://huggingface.co/docs/accelerate/v0.17.1/en/usage_guides/big_modeling
        checkpoint_files = None
        index_filename = None
        if os.path.isfile(checkpoint):
            if str(checkpoint).endswith(".json"):
                index_filename = checkpoint
            else:
                checkpoint_files = [checkpoint]
        elif os.path.isdir(checkpoint):
            # check if the whole state dict is present
            potential_state_bin = [
                f for f in os.listdir(checkpoint) if f == WEIGHTS_NAME
            ]
            potential_state_safetensor = [
                f for f in os.listdir(checkpoint) if f == SAFE_WEIGHTS_NAME
            ]
            if len(potential_state_bin) == 1:
                checkpoint_files = [os.path.join(checkpoint, potential_state_bin[0])]
            elif len(potential_state_safetensor) == 1:
                checkpoint_files = [
                    os.path.join(checkpoint, potential_state_safetensor[0])
                ]
            else:
                # otherwise check for sharded checkpoints
                potential_index = [
                    f for f in os.listdir(checkpoint) if f.endswith(".index.json")
                ]
                if len(potential_index) == 0:
                    raise ValueError(
                        f"{checkpoint} is not a folder containing a `.index.json` file or a {WEIGHTS_NAME} or a {SAFE_WEIGHTS_NAME} file"
                    )
                elif len(potential_index) == 1:
                    index_filename = os.path.join(checkpoint, potential_index[0])
                else:
                    raise ValueError(
                        f"{checkpoint} containing more than one `.index.json` file, delete the irrelevant ones."
                    )
        else:
            raise ValueError(
                "`checkpoint` should be the path to a file containing a whole state dict, or the index of a sharded "
                f"checkpoint, or a folder containing a sharded checkpoint or the whole state dict, but got {checkpoint}."
            )

        if index_filename is not None:
            checkpoint_folder = os.path.split(index_filename)[0]
            with open(index_filename) as f:
                index = json.loads(f.read())

            if "weight_map" in index:
                index = index["weight_map"]
            checkpoint_files = sorted(list(set(index.values())))
            checkpoint_files = [
                os.path.join(checkpoint_folder, f) for f in checkpoint_files
            ]
        else:
            index = None
        return index, index_filename, checkpoint_files

    def _load_tensor_from_checkpoint_file(
        self, checkpoint_file: str, key: str, update_cache: bool = True
    ) -> torch.Tensor:
        if checkpoint_file.endswith(".safetensors"):
            with safe_open(checkpoint_file, framework="pt", device=self._device) as f:
                tensor = f.get_tensor(key)
                if self._torch_dtype is not None:
                    tensor = tensor.to(self._torch_dtype)
                if update_cache and self._state_dict_cache is not None:
                    self._state_dict_cache[key] = tensor
                return tensor
        else:
            state_dict = torch.load(checkpoint_file, map_location=self._device)
            if update_cache:
                if self._state_dict_cache is not None:
                    self._state_dict_cache.update(state_dict)
                else:
                    log.warning(
                        f"Load full state dict from file {checkpoint_file}, but state dict cache is disabled."
                    )
            return state_dict[key]

    def __getitem__(self, key: str) -> torch.Tensor:
        if self._state_dict_cache is not None and key in self._state_dict_cache:
            return self._state_dict_cache[key]

        if self._index is None:
            if len(self._checkpoint_files) == 1 and os.path.isfile(
                self._checkpoint_files[0]
            ):
                checkpoint_file = self._checkpoint_files[0]
                tensor = self._load_tensor_from_checkpoint_file(
                    checkpoint_file, key, update_cache=True
                )
                return tensor
            else:
                if len(self._checkpoint_files) > 1:
                    raise RuntimeError(
                        "Get multiple checkpoint files, but index is not provided."
                    )
                if not os.path.isfile(self._checkpoint_files[0]):
                    raise FileNotFoundError(
                        f"Checkpoint file {self._checkpoint_files[0]} not found."
                    )
                raise RuntimeError("Unexpected error.")
        else:
            if key not in self._index:
                raise KeyError(f"Key {key} not found in index.")
            checkpoint_file = os.path.join(self._local_path, self._index[key])
            if not os.path.isfile(checkpoint_file):
                raise FileNotFoundError(f"Checkpoint file {checkpoint_file} not found.")
            tensor = self._load_tensor_from_checkpoint_file(
                checkpoint_file, key, update_cache=True
            )
            return tensor

    def __setitem__(self, key: str, value: torch.Tensor) -> None:
        """
        Set a tensor in the LazyStateDict. This will update the state dict cache if it is enabled.
        """
        assert key in list(
            self.keys()
        ), "KeyError: Cannot set a tensor for a key that does not exist in the LazyStateDict."
        if self._state_dict_cache is not None:
            self._state_dict_cache[key] = value
        else:
            log.warning(
                "State dict cache is disabled, setting a tensor will not update the cache."
            )
            self._state_dict_cache = {key: value}

    def __contains__(self, key: str) -> bool:
        if self._state_dict_cache is not None and key in self._state_dict_cache:
            return True
        if self._index is not None and key in self._index:
            return True
        if len(self._checkpoint_files) == 1 and os.path.isfile(
            self._checkpoint_files[0]
        ):
            try:
                tensor = self._load_tensor_from_checkpoint_file(
                    self._checkpoint_files[0], key, update_cache=False
                )
                return tensor is not None
            except Exception:
                return False
        return False

    def __len__(self) -> int:
        if self._index is not None:
            return len(self._index)
        if len(self._checkpoint_files) == 1 and os.path.isfile(
            self._checkpoint_files[0]
        ):
            checkpoint_file = self._checkpoint_files[0]
            if checkpoint_file.endswith(".safetensors"):
                with safe_open(checkpoint_file, framework="pt", device="cpu") as f:
                    return len(tuple(f.keys()))
            else:
                return len(
                    tuple(torch.load(checkpoint_file, map_location="cpu").keys())
                )
        raise RuntimeError(
            "Unexpected error: cannot determine the number of keys in the state dict."
        )

    def __iter__(self) -> Iterator[str]:
        if self._index is not None:
            return iter(self._index)
        elif self._state_dict_cache is not None:
            return iter(self._state_dict_cache)
        else:
            raise RuntimeError(
                "Unexpected error: cannot determine the keys in the state dict."
            )

    def keys(self) -> Iterator[str]:
        for key in self:
            yield key

    def values(self) -> Iterator[torch.Tensor]:
        for key in self:
            yield self[key]

    def items(self) -> Iterator[Tuple[str, torch.Tensor]]:
        for key in self:
            yield key, self[key]

    def __repr__(self) -> str:
        if self._index is not None:
            return f"{self.__class__.__name__}(keys={list(self.keys())})"
        else:
            return (
                f"{self.__class__.__name__}(checkpoint_files={self._checkpoint_files})"
            )

    def get_parameter(self, target: str) -> torch.Tensor:
        return self[target]

    def get_submodule(self, target: str) -> nn.Module:
        if self.meta_module is not None:
            module: nn.Module = deepcopy(self.meta_module.get_submodule(target))
            module.to_empty(device=self._device)
            state_dict = {}
            for name, _ in module.named_parameters():
                state_dict[name] = self[f"{target}.{name}"]
            module.load_state_dict(state_dict)
            return module
        else:
            raise RuntimeError(
                "Cannot get submodule because meta_module is not provided."
            )

    def load_state_dict(
        self, state_dict: Dict[str, torch.Tensor], strict: bool = True
    ) -> None:
        """
        Load a state dict into this LazyStateDict.
        This method is only for compatibility with nn.Module and it overrides the cache of LazyStateDict.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state dict to load.
            strict (bool): Whether to enforce that all keys in the state dict are present in this LazyStateDict.
        """
        log.warning(
            "Loading state dict into LazyStateDict is not recommended, as it may lead to unexpected behavior. "
            "Use with caution."
        )
        if strict:
            for key in state_dict:
                if key not in self:
                    raise KeyError(f"Key {key} not found in LazyStateDict.")
        for key, value in state_dict.items():
            self[key] = value



R"""
```bash
fusion_bench \
    modelpool=CausalLMPool/mixtral-8x7b \
    ...
```

if use flash attention 2, pass the following to the command line:

```bash
+modelpool.models._pretrained_.attn_implementation=flash_attention_2
```
"""

from .dynamic_skipping import DynamicSkippingPruningForMixtral
from .layer_wise_pruning import LayerWisePruningForMixtral
from .progressive_pruning import ProgressivePruningForMixtral

__all__ = [
    "DynamicSkippingPruningForMixtral",
    "LayerWisePruningForMixtral",
    "ProgressivePruningForMixtral",
]



classnames = [
    "banded",
    "blotchy",
    "braided",
    "bubbly",
    "bumpy",
    "chequered",
    "cobwebbed",
    "cracked",
    "crosshatched",
    "crystalline",
    "dotted",
    "fibrous",
    "flecked",
    "freckled",
    "frilly",
    "gauzy",
    "grid",
    "grooved",
    "honeycombed",
    "interlaced",
    "knitted",
    "lacelike",
    "lined",
    "marbled",
    "matted",
    "meshed",
    "paisley",
    "perforated",
    "pitted",
    "pleated",
    "polka-dotted",
    "porous",
    "potholed",
    "scaly",
    "smeared",
    "spiralled",
    "sprinkled",
    "stained",
    "stratified",
    "striped",
    "studded",
    "swirly",
    "veined",
    "waffled",
    "woven",
    "wrinkled",
    "zigzagged",
]

templates = [
    lambda c: f"a photo of a {c} texture.",
    lambda c: f"a photo of a {c} pattern.",
    lambda c: f"a photo of a {c} thing.",
    lambda c: f"a photo of a {c} object.",
    lambda c: f"a photo of the {c} texture.",
    lambda c: f"a photo of the {c} pattern.",
    lambda c: f"a photo of the {c} thing.",
    lambda c: f"a photo of the {c} object.",
]



"""
Original repo: https://github.com/Lucky-Lance/Expert_Sparsity

Reference:
    Not All Experts are Equal: Efficient Expert Pruning and Skipping for Mixture-of-Experts Large Language Models.
    ACL 2024.
    http://arxiv.org/abs/2402.14800
"""

from .mixtral import *



import os
from typing import TYPE_CHECKING, Optional

from datasets import Dataset, load_dataset, load_from_disk
from lightning.fabric.utilities import rank_zero_only
from tqdm.auto import tqdm

from fusion_bench.utils import timeit_context

from .alpaca import convert_alpaca_to_conversation

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer


def load_tokenized_metamathqa(
    tokenizer: "PreTrainedTokenizer",
    path: str = "meta-math/MetaMathQA",
    split: str = "train",
    cache_path: Optional[str] = None,
):
    if cache_path is not None and os.path.exists(cache_path):
        dataset = load_from_disk(cache_path)
        if split is not None and split in dataset:
            return dataset[split]
        else:
            return dataset

    dataset = load_dataset(path, split=split)

    # convert dataset to alpaca format and save to ../data/MetaMathQA.json
    alpaca_dataset = []
    for example in tqdm(dataset, disable=not rank_zero_only.rank == 0):
        alpaca_example = {
            "instruction": example["query"],
            "input": "",
            "output": example["response"],
        }
        alpaca_dataset.append(alpaca_example)

    conversations = convert_alpaca_to_conversation(alpaca_dataset)
    with timeit_context("Tokenizing dataset"):
        tokenized_dataset = tokenizer.apply_chat_template(
            conversations, return_dict=True
        )
    tokenized_dataset = Dataset.from_dict(tokenized_dataset)

    if cache_path is not None and rank_zero_only.rank == 0:
        tokenized_dataset.save_to_disk(cache_path)
    return tokenized_dataset



# flake8: noqa F401
import sys
from typing import TYPE_CHECKING

from fusion_bench.utils.lazy_imports import LazyImporter

_import_structure = {
    # --------------
    "base_algorithm": ["BaseModelFusionAlgorithm", "BaseAlgorithm"],
    "dummy": ["DummyAlgorithm"],
    # single task learning (fine-tuning)
    "classification": [
        "ImageClassificationFineTuningForCLIP",
        "ContinualImageClassificationFineTuningForCLIP",
    ],
    "lm_finetune": ["FullFinetuneSFT", "PeftFinetuneSFT", "BradleyTerryRewardModeling"],
    # analysis
    "analysis": ["TaskVectorCosSimilarity", "TaskVectorViolinPlot"],
    # model ensemble methods
    "ensemble": [
        "SimpleEnsembleAlgorithm",
        "WeightedEnsembleAlgorithm",
        "MaxModelPredictorAlgorithm",
    ],
    # model merging methods
    "linear": [
        "ExPOAlgorithm",
        "ExPOAlgorithmForLlama",
        "SimpleAverageForLlama",
        "TaskArithmeticForLlama",
        "LinearInterpolationAlgorithm",
    ],
    "slerp": ["SlerpMergeAlgorithm"],
    "simple_average": ["SimpleAverageAlgorithm"],
    "weighted_average": ["WeightedAverageAlgorithm", "WeightedAverageForLLama"],
    "task_arithmetic": ["TaskArithmeticAlgorithm"],
    "ties_merging": ["TiesMergingAlgorithm"],
    "dare": ["DareSimpleAverage", "DareTaskArithmetic", "DareTiesMerging"],
    "fisher_merging": [
        "FisherMergingForCLIPVisionModel",
        "FisherMergingAlgorithmForGPT2",
    ],
    "regmean": ["RegMeanAlgorithmForCLIP", "RegMeanAlgorithmForGPT2"],
    "regmean_plusplus": ["RegMeanAlgorithmForCLIPPlusPlus"],
    "adamerging": [
        "CLIPTaskWiseAdaMergingAlgorithm",
        "CLIPLayerWiseAdaMergingAlgorithm",
        "GPT2LayerWiseAdaMergingAlgorithm",
        "LayerWiseAdaMergingForLlamaSFT",
        "FlanT5LayerWiseAdaMergingAlgorithm",
    ],
    "pwe_moe": [
        "PWEMoELinearScalarizationForCLIP",
        "PWEMoExactParetoOptimalForCLIP",
    ],
    "ada_svd": ["AdaSVDMergingForCLIPVisionModel"],
    "doge_ta": ["DOGE_TA_Algorithm"],
    "task_singular_vector": ["TaskSingularVectorMerging"],
    "isotropic_merging": [
        "ISO_C_Merge",  # alias
        "ISO_CTS_Merge",  # alias
        "IsotropicMergingInCommonAndTaskSubspace",
        "IsotropicMergingInCommonSubspace",
    ],
    "opcm": ["OPCMForCLIP"],
    "gossip": [
        "CLIPLayerWiseGossipAlgorithm",
        "CLIPTaskWiseGossipAlgorithm",
        "FlanT5LayerWiseGossipAlgorithm",
    ],
    "fw_merging": ["FrankWolfeHardAlgorithm", "FrankWolfeSoftAlgorithm"],
    # plug-and-play model merging methods
    "concrete_subspace": [
        "ConcreteTaskArithmeticAlgorithmForCLIP",
        "ConcreteTaskWiseAdaMergingForCLIP",
        "ConcreteLayerWiseAdaMergingForCLIP",
        "ConcreteSafeLayerWiseAdaMergingForCLIP",
        "ConcreteSafeTaskWiseAdaMergingForCLIP",
        "PostDefenseAWMAlgorithmForCLIP",
        "PostDefenseSAUAlgorithmForCLIP",
    ],
    # model mixing methods
    "depth_upscaling": ["DepthUpscalingAlgorithm", "DepthUpscalingForLlama"],
    "mixture_of_experts": [
        "MixtralUpscalingAlgorithm",
        "MixtralForCausalLMUpscalingAlgorithm",
        "MixtralMoEMergingAlgorithm",
        "MixtralForCausalLMMergingAlgorithm",
    ],
    "dawe": ["DataAdaptiveWeightEnsemblingForCLIP"],
    "we_moe": ["CLIPWeightEnsemblingMoEAlgorithm"],
    "rankone_moe": ["CLIPRankOneMoEAlgorithm", "RankOneMoEAlgorithm"],
    "sparse_we_moe": [
        "SparseWeightEnsemblingMoEAlgorithm",
        "SparseCLIPWeightEnsemblingMoEAlgorithm",
    ],
    "model_recombination": ["ModelRecombinationAlgorithm"],
    "smile_upscaling": [
        "SmileUpscalingAlgorithm",
        "SingularProjectionMergingAlgorithm",
    ],
    # pruning methods
    "pruning": [
        "MagnitudeDiffPruningAlgorithm",
        "RandomPruningForLlama",
        "MagnitudePruningForLlama",
        "WandaPruningForLlama",
        "SparseGPTPruningForLlama",
    ],
    "sparselo": [
        "IterativeSparseLoForLlama",
        "SparseLoForLlama",
        "PCPSparseLoForLlama",
    ],
    # MoE expert pruning
    "expert_sparsity": [
        "DynamicSkippingPruningForMixtral",
        "LayerWisePruningForMixtral",
        "ProgressivePruningForMixtral",
    ],
}


if TYPE_CHECKING:
    from .ada_svd import AdaSVDMergingForCLIPVisionModel
    from .adamerging import *
    from .analysis import TaskVectorCosSimilarity, TaskVectorViolinPlot
    from .base_algorithm import BaseAlgorithm, BaseModelFusionAlgorithm
    from .classification import (
        ContinualImageClassificationFineTuningForCLIP,
        ImageClassificationFineTuningForCLIP,
    )
    from .concrete_subspace import (
        ConcreteLayerWiseAdaMergingForCLIP,
        ConcreteSafeLayerWiseAdaMergingForCLIP,
        ConcreteSafeTaskWiseAdaMergingForCLIP,
        ConcreteTaskArithmeticAlgorithmForCLIP,
        ConcreteTaskWiseAdaMergingForCLIP,
        PostDefenseAWMAlgorithmForCLIP,
        PostDefenseSAUAlgorithmForCLIP,
    )
    from .dare import DareSimpleAverage, DareTaskArithmetic, DareTiesMerging
    from .dawe import DataAdaptiveWeightEnsemblingForCLIP
    from .depth_upscaling import DepthUpscalingAlgorithm, DepthUpscalingForLlama
    from .doge_ta import DOGE_TA_Algorithm
    from .dummy import DummyAlgorithm
    from .ensemble import (
        MaxModelPredictorAlgorithm,
        SimpleEnsembleAlgorithm,
        WeightedEnsembleAlgorithm,
    )
    from .expert_sparsity import (
        DynamicSkippingPruningForMixtral,
        LayerWisePruningForMixtral,
        ProgressivePruningForMixtral,
    )
    from .fisher_merging import FisherMergingForCLIPVisionModel
    from .fw_merging import FrankWolfeHardAlgorithm, FrankWolfeSoftAlgorithm
    from .gossip import (
        CLIPLayerWiseGossipAlgorithm,
        CLIPTaskWiseGossipAlgorithm,
        FlanT5LayerWiseGossipAlgorithm,
    )
    from .isotropic_merging import (
        ISO_C_Merge,
        ISO_CTS_Merge,
        IsotropicMergingInCommonAndTaskSubspace,
        IsotropicMergingInCommonSubspace,
    )
    from .linear import (
        ExPOAlgorithm,
        ExPOAlgorithmForLlama,
        LinearInterpolationAlgorithm,
        SimpleAverageForLlama,
        TaskArithmeticForLlama,
    )
    from .lm_finetune import *
    from .mixture_of_experts import (
        MixtralForCausalLMMergingAlgorithm,
        MixtralForCausalLMUpscalingAlgorithm,
        MixtralMoEMergingAlgorithm,
        MixtralUpscalingAlgorithm,
    )
    from .model_recombination import ModelRecombinationAlgorithm
    from .opcm import OPCMForCLIP
    from .pruning import (
        MagnitudeDiffPruningAlgorithm,
        MagnitudePruningForLlama,
        RandomPruningForLlama,
        SparseGPTPruningForLlama,
        WandaPruningForLlama,
    )
    from .pwe_moe import (
        PWEMoELinearScalarizationForCLIP,
        PWEMoExactParetoOptimalForCLIP,
    )
    from .rankone_moe import CLIPRankOneMoEAlgorithm, RankOneMoEAlgorithm
    from .regmean import RegMeanAlgorithmForCLIP, RegMeanAlgorithmForGPT2
    from .regmean_plusplus import RegMeanAlgorithmForCLIPPlusPlus
    from .simple_average import SimpleAverageAlgorithm
    from .slerp import SlerpMergeAlgorithm
    from .smile_upscaling import (
        SingularProjectionMergingAlgorithm,
        SmileUpscalingAlgorithm,
    )
    from .sparse_we_moe import (
        SparseCLIPWeightEnsemblingMoEAlgorithm,
        SparseWeightEnsemblingMoEAlgorithm,
    )
    from .sparselo import (
        IterativeSparseLoForLlama,
        PCPSparseLoForLlama,
        SparseLoForLlama,
    )
    from .task_arithmetic import TaskArithmeticAlgorithm
    from .task_singular_vector import TaskSingularVectorMerging
    from .ties_merging import TiesMergingAlgorithm
    from .we_moe import CLIPWeightEnsemblingMoEAlgorithm
    from .weighted_average import WeightedAverageAlgorithm, WeightedAverageForLLama

else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
    )



import logging
import os
from typing import List, Literal, Optional, Union, TYPE_CHECKING

import lightning.fabric
import lm_eval
import lm_eval.models
from lm_eval.__main__ import check_argument_types, cli_evaluate, setup_parser
from omegaconf import DictConfig, ListConfig

from fusion_bench import BaseTaskPool
from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.utils.strenum import _version


log = logging.getLogger(__name__)


class LMEvalHarnessTaskPool(BaseTaskPool, LightningFabricMixin):
    def __init__(
        self,
        tasks: Union[str, List[str]],
        apply_chat_template: bool = False,
        include_path: Optional[str] = None,
        batch_size: int = 1,
        metadata: Optional[DictConfig] = None,
        verbosity: Optional[
            Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
        ] = None,
        output_path: Optional[str] = None,
        log_samples: bool = False,
        _usage_: Optional[str] = None,
        _version_: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(_usage_=_usage_, _version_=_version_)
        self.tasks = tasks
        self.include_path = include_path
        self.batch_size = batch_size
        self.metadata = metadata
        self.apply_chat_template = apply_chat_template
        self.verbosity = verbosity
        self.kwargs = kwargs
        self.output_path = output_path
        self.log_samples = log_samples

    def evaluate(self, model, *command_line_args, **kwargs):
        command_line_args = []
        if self.include_path is not None:
            command_line_args.extend(["--include_path", self.include_path])
        if isinstance(self.tasks, (list, ListConfig)):
            command_line_args.extend(["--tasks", ",".join(self.tasks)])
        elif isinstance(self.tasks, str):
            command_line_args.extend(["--tasks", self.tasks])
        if self.apply_chat_template:
            command_line_args.extend(
                ["--apply_chat_template", str(self.apply_chat_template)]
            )
        if self.batch_size is not None:
            command_line_args.extend(["--batch_size", str(self.batch_size)])
        if self.verbosity is not None:
            command_line_args.extend(["--verbosity", str(self.verbosity)])
        if self.metadata is not None:
            command_line_args.extend(["--metadata", str(self.metadata)])
        if self.output_path is None:
            command_line_args.extend(
                [
                    "--output_path",
                    os.path.join(self.log_dir, "lm_eval_results"),
                ]
            )
        else:
            command_line_args.extend(["--output_path", self.output_path])
        if self.log_samples:
            command_line_args.extend(["--log_samples"])
        for key, value in kwargs.items():
            command_line_args.extend([f"--{key}", str(value)])

        parser = setup_parser()
        check_argument_types(parser)
        args = parser.parse_args(args=command_line_args)
        log.info("LM-Eval Harness arguments:\n%s", args)

        if not lightning.fabric.is_wrapped(model):
            model = self.fabric.setup(model)
        args.model = lm_eval.models.huggingface.HFLM(pretrained=model)
        cli_evaluate(args)



"""
Implementation of Task Arithmetic in Trust Region: A Training-Free Model Merging Approach to Navigate Knowledge Conflicts
https://openreview.net/forum?id=q3ztjJRQuJ
"""

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, Iterable, List, Union

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from typing_extensions import override

from fusion_bench import BaseAlgorithm, BaseModelPool
from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import CLIPClassificationMixin, SimpleProfilerMixin
from fusion_bench.utils import first
from fusion_bench.utils.state_dict_arithmetic import state_dict_sub
from fusion_bench.utils.type import StateDictType

from .utils import state_dict_to_vector, vector_to_state_dict

log = logging.getLogger(__name__)


def trainable_state_dict(module: nn.Module) -> StateDictType:
    """
    Returns the state dictionary of the module containing only the trainable parameters.

    Args:
        module (nn.Module): The neural network module.

    Returns:
        Dict[str, Tensor]: A dictionary containing the names and values of the trainable parameters.
    """
    return {
        name: param for name, param in module.named_parameters() if param.requires_grad
    }


class TaskArithmeticWithTrustRegionForCLIP(
    BaseAlgorithm,
    SimpleProfilerMixin,
    CLIPClassificationMixin,
):
    def __init__(
        self,
        scaling_factor: Union[float, List[float]],
        threshold_quantile: float,
        max_samples: int,
        batch_size: int,
        zero_shot: bool,
        **kwargs,
    ):
        self.scaling_factor = scaling_factor
        self.threshold_quantile = threshold_quantile
        self.max_samples = max_samples
        self.batch_size = batch_size
        self.zero_shot = zero_shot
        super().__init__(**kwargs)

    @override
    def run(self, modelpool: BaseModelPool):
        self.modelpool = modelpool

        # compute the task vectors
        pretrained_model, task_vectors = self.compute_vanilla_task_vectors()
        task_vectors = {
            name: state_dict_to_vector(task_vector)
            for name, task_vector in task_vectors.items()
        }

        if not self.zero_shot:
            all_avg_abs_grads = self.compute_avg_abs_grads(pretrained_model)
            all_avg_abs_grads = {
                n: state_dict_to_vector(grad) for n, grad in all_avg_abs_grads.items()
            }
        else:
            # the task vector is used to estimate the gradient
            all_avg_abs_grads = {name: tv.abs() for name, tv in task_vectors.items()}

        # compute the trust region
        Omega = torch.zeros_like(first(all_avg_abs_grads.values()))

        for i in all_avg_abs_grads:
            for j in all_avg_abs_grads:
                if i != j:
                    vector1 = all_avg_abs_grads[i]
                    vector2 = torch.abs(task_vectors[j])
                    Omega += vector1 * vector2

        values, indices = Omega.sort(descending=False)
        threshold = values[
            max(0, min(int(Omega.numel() * self.threshold_quantile), Omega.numel() - 1))
        ]

        mask = (Omega < threshold).bool()

        # compute the task vectors
        for task in task_vectors:
            task_vectors[task] = task_vectors[task] * mask

        task_vector_sum = sum(task_vectors.values())
        task_vector_sum = vector_to_state_dict(
            task_vector_sum, trainable_state_dict(pretrained_model)
        )

        if isinstance(self.scaling_factor, (int, float)):
            model = pretrained_model
            for name, param in model.named_parameters():
                param.data += task_vector_sum[name] * self.scaling_factor
            return model
        elif isinstance(self.scaling_factor, Iterable):
            models = {}
            for scaling_factor in self.scaling_factor:
                model = deepcopy(pretrained_model)
                for name, param in pretrained_model.named_parameters():
                    param.data += task_vector_sum[name] * scaling_factor
                models[scaling_factor] = model
            return models
        else:
            raise ValueError(
                f"Incorrect type of `scaling_factor`: {type(self.scaling_factor)}. "
                "It should be a single real number or a list of real numbers."
            )

    def compute_avg_abs_grads(self, pretrained_model):
        modelpool = self.modelpool

        self.setup_zero_shot_classification_head()

        pretrained_model = (
            deepcopy(pretrained_model)
            if pretrained_model is not None
            else modelpool.load_pretrained_model()
        )
        pretrained_model = self.fabric.setup_module(pretrained_model)
        pretrained_model.train()

        all_avg_abs_grads: Dict[str, StateDictType] = {}
        for train_dataset_name in (
            pbar := tqdm(
                modelpool.train_dataset_names, desc="Train datasets", dynamic_ncols=True
            )
        ):
            pbar.set_description(f"Train dataset: {train_dataset_name}")
            dataset = modelpool.load_train_dataset(train_dataset_name)
            dataset = CLIPDataset(dataset, self.clip_processor)
            dataloader = DataLoader(dataset, shuffle=True, batch_size=self.batch_size)
            dataloader = self.fabric.setup_dataloaders(dataloader)

            grad: StateDictType = defaultdict(float)
            num_samples = 0
            for batch in dataloader:
                images, labels = batch
                batch_size = images.size(0)

                if num_samples + batch_size > self.max_samples:
                    batch_size = self.max_samples - num_samples
                    images = images[:batch_size]
                    labels = labels[:batch_size]

                logits = self.compute_logits(
                    pretrained_model, images, task=train_dataset_name
                )
                for i in range(batch_size):
                    pretrained_model.zero_grad()
                    loss = F.cross_entropy(logits[i], labels[i])
                    self.fabric.backward(
                        loss, retain_graph=True if i != batch_size - 1 else False
                    )
                    for name, param in pretrained_model.module.named_parameters():
                        if param.requires_grad:
                            grad[name] += torch.abs(param.grad).detach()

                num_samples += batch_size
                if num_samples >= self.max_samples:
                    break

            for name in grad:
                grad[name] = (grad[name] / num_samples).cpu()

            all_avg_abs_grads[name] = grad
        return all_avg_abs_grads

    @torch.no_grad()
    def compute_vanilla_task_vectors(self):
        modelpool = self.modelpool

        pretrained_model = modelpool.load_pretrained_model()
        pretrained_sd = trainable_state_dict(pretrained_model)
        finetuned_sds = {
            name: trainable_state_dict(model)
            for name, model in modelpool.named_models()
        }

        task_vectors = {
            name: state_dict_sub(finetuned, pretrained_sd)
            for name, finetuned in finetuned_sds.items()
        }
        return pretrained_model, task_vectors



import functools
import inspect
from typing import Callable, Union

from .packages import import_object


@functools.cache
def cached_func_call(func: Union[str, Callable], *args, **kwargs):
    R"""
    Calls the given function with the provided arguments and caches the result.

    This function uses `functools.cache` to cache the result of the function call,
    so that subsequent calls with the same arguments return the cached result
    instead of recomputing it.

    Args:
        func (Union[str, Callable]): The function to be called. If `func` is a string, it is assumed to be the absolute name of the function to be imported.
        *args: Variable length argument list to be passed to the function.
        **kwargs: Arbitrary keyword arguments to be passed to the function.

    Returns:
        The result of the function call.
    """
    if isinstance(func, str):
        func = import_object(func)
    return func(*args, **kwargs)


def number_of_arguments(func):
    """
    Return the number of arguments of the passed function, even if it's a partial function.
    """
    if isinstance(func, functools.partial):
        total_args = len(inspect.signature(func.func).parameters)
        return total_args - len(func.args) - len(func.keywords)
    return len(inspect.signature(func).parameters)



# flake8: noqa F401
import sys

from typing_extensions import TYPE_CHECKING

from fusion_bench.utils.lazy_imports import LazyImporter

_import_structure = {
    "clip_classification": ["CLIPClassificationMixin"],
    "fabric_training": ["FabricTrainingMixin"],
    "hydra_config": ["HydraConfigMixin"],
    "lightning_fabric": ["LightningFabricMixin"],
    "openclip_classification": ["OpenCLIPClassificationMixin"],
    "serialization": ["YAMLSerializationMixin", "BaseYAMLSerializableModel"],
    "simple_profiler": ["SimpleProfilerMixin"],
}

if TYPE_CHECKING:
    from .clip_classification import CLIPClassificationMixin
    from .fabric_training import FabricTrainingMixin
    from .hydra_config import HydraConfigMixin
    from .lightning_fabric import LightningFabricMixin
    from .openclip_classification import OpenCLIPClassificationMixin
    from .serialization import BaseYAMLSerializableModel, YAMLSerializationMixin
    from .simple_profiler import SimpleProfilerMixin
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
    )



import json
import logging
import os
from typing import Callable, Dict, Iterable, Optional, Union  # noqa: F401

import lightning as L
from lightning.fabric.utilities.rank_zero import rank_zero_only
from omegaconf import DictConfig, OmegaConf
from torch import nn
from tqdm.auto import tqdm

import fusion_bench.utils.instantiate_utils
from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.programs import BaseHydraProgram
from fusion_bench.taskpool import BaseTaskPool
from fusion_bench.utils import import_object, instantiate, timeit_context
from fusion_bench.utils.hydra_utils import get_hydra_output_dir
from fusion_bench.utils.json import print_json
from fusion_bench.utils.rich_utils import print_bordered, print_config_tree
from fusion_bench.utils.pylogger import getRankZeroLogger

log = getRankZeroLogger(__name__)


class FabricModelFusionProgram(
    LightningFabricMixin,
    BaseHydraProgram,
):
    method: BaseAlgorithm
    modelpool: BaseModelPool
    taskpool: Optional[BaseTaskPool] = None

    _config_mapping = BaseHydraProgram._config_mapping | {
        "_method": "method",
        "_modelpool": "modelpool",
        "_taskpool": "taskpool",
        "_fabric": "fabric",
        "fast_dev_run": "fast_dev_run",
        "seed": "seed",
    }

    def __init__(
        self,
        method: DictConfig,
        modelpool: DictConfig,
        taskpool: Optional[DictConfig] = None,
        *,
        fabric: Optional[DictConfig] = None,
        print_config: bool = True,
        dry_run: bool = False,
        report_save_path: Optional[str] = None,
        merged_model_save_path: Optional[str] = None,
        merged_model_save_kwargs: Optional[DictConfig] = None,
        fast_dev_run: bool = False,
        seed: Optional[int] = None,
        print_function_call: bool = True,
        **kwargs,
    ):
        self._method = method
        self._modelpool = modelpool
        self._taskpool = taskpool
        self._fabric = fabric
        self.report_save_path = report_save_path
        self.merged_model_save_path = merged_model_save_path
        self.merged_model_save_kwargs = merged_model_save_kwargs
        self.fast_dev_run = fast_dev_run
        self.seed = seed
        fusion_bench.utils.instantiate_utils.PRINT_FUNCTION_CALL = print_function_call
        super().__init__(**kwargs)

        if print_config:
            print_config_tree(
                self.config,
                print_order=["method", "modelpool", "taskpool"],
            )
        if dry_run:
            log.info("The program is running in dry-run mode. Exiting.")
            exit(0)

    def _instantiate_and_setup(
        self, config: DictConfig, compat_load_fn: Optional[str] = None
    ):
        R"""
        Instantiates and sets up an object based on the provided configuration.

        This method performs the following steps:
        1. Checks if the configuration dictionary contains the key "_target_".
        2. If "_target_" is not found (for v0.1.x), attempts to instantiate the object using a compatible load function if provided.
           - Logs a warning if "_target_" is missing.
           - If `compat_load_fn` is provided, imports the function and uses it to instantiate the object.
           - If `compat_load_fn` is not provided, raises a ValueError.
        3. If "_target_" is found (for v.0.2.0 and above), attempts to import and instantiate the object using the `instantiate` function.
           - Ensures the target can be imported.
           - Uses the `instantiate` function with `_recursive_` set based on the configuration.
        4. Sets the `_program` attribute of the instantiated object to `self` if the object has this attribute.
        5. Sets the `_fabric_instance` attribute of the instantiated object to `self.fabric` if the object has this attribute and `self.fabric` is not None.
        6. Returns the instantiated and set up object.
        """
        if "_target_" not in config:
            log.warning(
                "No '_target_' key found in config. Attempting to instantiate the object in a compatible way."
            )
            if compat_load_fn is not None:
                compat_load_fn = import_object(compat_load_fn)
                if rank_zero_only.rank == 0:
                    print_bordered(
                        OmegaConf.to_yaml(config),
                        title="instantiate compat object",
                        style="magenta",
                        code_style="yaml",
                    )
                obj = compat_load_fn(config)
            else:
                raise ValueError(
                    "No load function provided. Please provide a load function to instantiate the object."
                )
        else:
            # try to import the object from the target
            # this checks if the target is valid and can be imported
            import_object(config._target_)
            obj = instantiate(
                config,
                _recursive_=config.get("_recursive_", False),
            )
        if hasattr(obj, "_program"):
            obj._program = self
        if hasattr(obj, "_fabric_instance") and self.fabric is not None:
            obj._fabric_instance = self.fabric
        if hasattr(obj, "_fabric") and self.fabric is not None:
            # for v0.1.x compatibility
            obj._fabric = self.fabric
        return obj

    def save_merged_model(self, merged_model):
        """
        Saves the merged model to the specified path.
        """
        if self.merged_model_save_path is not None:
            # path to save the merged model, use "{log_dir}" to refer to the logger directory
            save_path: str = self.merged_model_save_path
            if "{log_dir}" in save_path and self.log_dir is not None:
                save_path = save_path.format(log_dir=self.log_dir)

            if os.path.dirname(save_path):
                os.makedirs(os.path.dirname(save_path), exist_ok=True)

            # save the merged model
            if self.merged_model_save_kwargs is not None:
                merged_model_save_kwargs = self.merged_model_save_kwargs
            else:
                merged_model_save_kwargs = {}
            with timeit_context(f"Saving the merged model to {save_path}"):
                self.modelpool.save_model(
                    merged_model,
                    save_path,
                    **merged_model_save_kwargs,
                )
        else:
            print("No save path specified for the merged model. Skipping saving.")

    def evaluate_merged_model(
        self,
        taskpool: BaseTaskPool,
        merged_model: Union[nn.Module, Dict, Iterable],
        *args,
        **kwargs,
    ):
        """
        Evaluates the merged model using the provided task pool.

        Depending on the type of the merged model, this function handles the evaluation differently:
        - If the merged model is an instance of `nn.Module`, it directly evaluates the model.
        - If the merged model is a dictionary, it extracts the model from the dictionary and evaluates it.
          The evaluation report is then updated with the remaining dictionary items.
        - If the merged model is an iterable, it recursively evaluates each model in the iterable.
        - Raises a `ValueError` if the merged model is of an invalid type.

        Args:
            taskpool: The task pool used for evaluating the merged model.
            merged_model: The merged model to be evaluated. It can be an instance of `nn.Module`, a dictionary, or an iterable.
            *args: Additional positional arguments to be passed to the `evaluate` method of the taskpool.
            **kwargs: Additional keyword arguments to be passed to the `evaluate` method of the taskpool.

        Returns:
            The evaluation report. The type of the report depends on the type of the merged model:
            - If the merged model is an instance of `nn.Module`, the report is a dictionary.
            - If the merged model is a dictionary, the report is a dictionary updated with the remaining dictionary items.
            - If the merged model is an iterable, the report is a list of evaluation reports.
        """
        if isinstance(merged_model, nn.Module):
            report = taskpool.evaluate(merged_model, *args, **kwargs)
            return report
        elif isinstance(merged_model, Dict):
            report = {}
            for key, item in merged_model.items():
                if isinstance(item, nn.Module):
                    report[key] = taskpool.evaluate(item, *args, **kwargs)
                elif key == "models":
                    # for multi-model evaluation
                    report[key] = self.evaluate_merged_model(
                        taskpool, item, *args, **kwargs
                    )
                else:
                    # metadata
                    report[key] = item
            return report
        elif isinstance(merged_model, Iterable):
            return [
                self.evaluate_merged_model(taskpool, m, *args, **kwargs)
                for m in tqdm(merged_model, desc="Evaluating models")
            ]
        else:
            raise ValueError(f"Invalid type for merged model: {type(merged_model)}")

    def run(self):
        """
        Executes the model fusion program.
        """
        fabric = self.fabric
        if self.seed is not None:
            L.seed_everything(self.seed)
        if fabric.global_rank == 0:
            self._link_hydra_output()

        log.info("Running the model fusion program.")
        # setup the modelpool, method, and taskpool
        log.info("loading model pool")
        self.modelpool = self._instantiate_and_setup(
            self._modelpool,
            compat_load_fn="fusion_bench.compat.modelpool.load_modelpool_from_config",
        )
        log.info("loading method")
        self.method = self._instantiate_and_setup(
            self._method,
            compat_load_fn="fusion_bench.compat.method.load_algorithm_from_config",
        )
        if self._taskpool is not None:
            log.info("loading task pool")
            self.taskpool = self._instantiate_and_setup(
                self._taskpool,
                compat_load_fn="fusion_bench.compat.taskpool.load_taskpool_from_config",
            )

        merged_model = self.method.run(self.modelpool)
        if merged_model is None:
            log.info(
                "No merged model returned by the method. Skipping saving and evaluation."
            )
        else:
            self.save_merged_model(merged_model)
            if self.taskpool is not None:
                report = self.evaluate_merged_model(self.taskpool, merged_model)
                try:
                    if rank_zero_only.rank == 0:
                        print_json(report, print_type=False)
                except Exception as e:
                    log.warning(f"Failed to pretty print the report: {e}")
                    log.info(report)
                if self.report_save_path is not None:
                    # save report (Dict) to a file
                    # if the directory of `save_report` does not exists, create it
                    if "{log_dir}" in self.report_save_path and self.log_dir is not None:
                        self.report_save_path = self.report_save_path.format(log_dir=self.log_dir)
                    os.makedirs(os.path.dirname(self.report_save_path), exist_ok=True)
                    json.dump(report, open(self.report_save_path, "w"))
            else:
                log.info("No task pool specified. Skipping evaluation.")

    @rank_zero_only
    def _link_hydra_output(self):
        """
        Creates a symbolic link to the Hydra output directory within the specified log directory.

        If `self.log_dir` is not None, this method will:
        1. Retrieve the Hydra output directory using `get_hydra_output_dir()`.
        2. Create the log directory if it does not already exist.
        3. Create a symbolic link named "hydra_output_<basename_of_hydra_output_dir>"
           within the log directory, pointing to the Hydra output directory.

        Note:
            - The symbolic link is created only if the Hydra output directory is not None.
            - The `target_is_directory` parameter is set to True to indicate that the target is a directory.

        Raises:
            OSError: If the symbolic link creation fails.
        """
        if self.log_dir is not None:
            # make symlink to the hydra output directory
            try:
                hydra_output_dir = get_hydra_output_dir()
            except Exception as e:
                hydra_output_dir = None

            if hydra_output_dir is not None:
                os.makedirs(self.log_dir, exist_ok=True)
                try:
                    # if the system is windows, use the `mklink` command in "CMD" to create the symlink
                    if os.name == "nt":
                        os.system(f"mklink /J {os.path.abspath(os.path.join(self.log_dir, 'hydra_output_' + os.path.basename(hydra_output_dir)))} {os.path.abspath(hydra_output_dir)}")
                    else:
                        os.symlink(
                            hydra_output_dir,
                            os.path.join(
                                self.log_dir,
                                "hydra_output_" + os.path.basename(hydra_output_dir),
                            ),
                            target_is_directory=True,
                        )
                except OSError as e:
                    log.warning(f"Failed to create symbolic link: {e}")



classnames = [
    "airplane",
    "bird",
    "car",
    "cat",
    "deer",
    "dog",
    "horse",
    "monkey",
    "ship",
    "truck",
]

templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
]



R"""
Overview of Ties-Merging:

1. Trim: For each task t, we trim the redundant parameters from the task vector $\tau_t$ to create $\hat{\tau}_t$ by keeping the top-k% values according to their magnitude and trimming the bottom $(100 - k)\%$ of the redundant parameters by resetting them to 0. This can be decomposed further as $\hat{\tau}_t = \hat{\gamma}_t \odot \hat{\mu}_t$.

2. Elect: Next, we create an aggregate elected sign vector $\gamma_m$ for the merged model that resolves the disagreements in the sign for each parameter p across different models. To create the elected sign vector, we choose the sign with the highest total magnitude across all relevant models. For each parameter $p \in \{1, 2, \ldots, d\}$, we separate the values $\{\hat{\tau}_t^p\}_{t=1}^n$ based on their sign $(+1$ or $-1)$ and take their sum to calculate the total mass (i.e., total magnitude) in the positive and the negative direction. We then assign $\gamma_m^p$ as the sign with greater total movement. This can be efficiently computed using $\gamma_m^p = \text{sgn}(\sum_{t=1}^n \hat{\tau}_t^p)$.

3. Disjoint Merge: Then, for each parameter p, we compute a disjoint mean by only keeping the parameter values from the models whose signs are the same as the aggregated elected sign and calculate their mean. Formally, let $A_p = \{t \in [n] \mid \hat{\gamma}_t^p = \gamma_m^p\}$, then $\tau_m^p = \frac{1}{|A_p|}\sum_{t\in A_p} \hat{\tau}_t^p$. Note that the disjoint mean always ignores the zero values.
"""

import logging
from typing import Dict, List, Literal, Mapping, Union  # noqa: F401

import torch
from torch import Tensor, nn

from fusion_bench.compat.modelpool import to_modelpool
from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.type import StateDictType

from .ties_merging_utils import state_dict_to_vector, ties_merging, vector_to_state_dict

log = logging.getLogger(__name__)


class TiesMergingAlgorithm(BaseAlgorithm, SimpleProfilerMixin):
    """
    TiesMergingAlgorithm is a class for fusing multiple models using the TIES merging technique.

    Attributes:
        scaling_factor (float): The scaling factor to apply to the merged task vector.
        threshold (float): The threshold for resetting values in the task vector.
        remove_keys (List[str]): List of keys to remove from the state dictionary.
        merge_func (Literal["sum", "mean", "max"]): The merge function to use for disjoint merging.
    """

    _config_mapping = BaseAlgorithm._config_mapping | {
        "scaling_factor": "scaling_factor",
        "threshold": "threshold",
        "remove_keys": "remove_keys",
        "merge_func": "merge_func",
    }

    def __init__(
        self,
        scaling_factor: float,
        threshold: float,
        remove_keys: List[str],
        merge_func: Literal["sum", "mean", "max"],
        **kwargs,
    ):
        """
        Initialize the TiesMergingAlgorithm with the given parameters.

        Args:
            scaling_factor (float): The scaling factor to apply to the merged task vector.
            threshold (float): The threshold for resetting values in the task vector.
            remove_keys (List[str]): List of keys to remove from the state dictionary.
            merge_func (Literal["sum", "mean", "max"]): The merge function to use for disjoint merging.
            **kwargs: Additional keyword arguments for the base class.
        """
        self.scaling_factor = scaling_factor
        self.threshold = threshold
        self.remove_keys = remove_keys
        self.merge_func = merge_func
        super().__init__(**kwargs)

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool | Dict[str, nn.Module], **kwargs):
        """
        Run the TIES merging algorithm to fuse models in the model pool.

        Args:
            modelpool (BaseModelPool | Dict[str, nn.Module]): The model pool containing the models to fuse.

        Returns:
            nn.Module: The fused model.
        """
        log.info("Fusing models using ties merging.")
        modelpool = to_modelpool(modelpool)
        remove_keys = self.config.get("remove_keys", [])
        merge_func = self.config.get("merge_func", "sum")
        scaling_factor = self.scaling_factor
        threshold = self.threshold

        with self.profile("loading models"):
            # Load the pretrained model
            pretrained_model = modelpool.load_model("_pretrained_")

            # Load the state dicts of the models
            ft_checks: List[StateDictType] = [
                modelpool.load_model(model_name).state_dict(keep_vars=True)
                for model_name in modelpool.model_names
            ]
            ptm_check: StateDictType = pretrained_model.state_dict(keep_vars=True)

        with self.profile("merging models"):
            # Compute the task vectors
            flat_ft: Tensor = torch.vstack(
                [state_dict_to_vector(check, remove_keys) for check in ft_checks]
            )
            flat_ptm: Tensor = state_dict_to_vector(ptm_check, remove_keys)
            tv_flat_checks = flat_ft - flat_ptm

            # Perform TIES Merging
            merged_tv = ties_merging(
                tv_flat_checks,
                reset_thresh=threshold,
                merge_func=merge_func,
            )
            merged_check = flat_ptm + scaling_factor * merged_tv
            merged_state_dict = vector_to_state_dict(
                merged_check, ptm_check, remove_keys=remove_keys
            )

            # Load the merged state dict into the pretrained model
            pretrained_model.load_state_dict(merged_state_dict)

        self.print_profile_summary()
        return pretrained_model



classname_mapping = {
    "0": "Black_footed_Albatross",
    "1": "Laysan_Albatross",
    "2": "Sooty_Albatross",
    "3": "Groove_billed_Ani",
    "4": "Crested_Auklet",
    "5": "Least_Auklet",
    "6": "Parakeet_Auklet",
    "7": "Rhinoceros_Auklet",
    "8": "Brewer_Blackbird",
    "9": "Red_winged_Blackbird",
    "10": "Rusty_Blackbird",
    "11": "Yellow_headed_Blackbird",
    "12": "Bobolink",
    "13": "Indigo_Bunting",
    "14": "Lazuli_Bunting",
    "15": "Painted_Bunting",
    "16": "Cardinal",
    "17": "Spotted_Catbird",
    "18": "Gray_Catbird",
    "19": "Yellow_breasted_Chat",
    "20": "Eastern_Towhee",
    "21": "Chuck_will_Widow",
    "22": "Brandt_Cormorant",
    "23": "Red_faced_Cormorant",
    "24": "Pelagic_Cormorant",
    "25": "Bronzed_Cowbird",
    "26": "Shiny_Cowbird",
    "27": "Brown_Creeper",
    "28": "American_Crow",
    "29": "Fish_Crow",
    "30": "Black_billed_Cuckoo",
    "31": "Mangrove_Cuckoo",
    "32": "Yellow_billed_Cuckoo",
    "33": "Gray_crowned_Rosy_Finch",
    "34": "Purple_Finch",
    "35": "Northern_Flicker",
    "36": "Acadian_Flycatcher",
    "37": "Great_Crested_Flycatcher",
    "38": "Least_Flycatcher",
    "39": "Olive_sided_Flycatcher",
    "40": "Scissor_tailed_Flycatcher",
    "41": "Vermilion_Flycatcher",
    "42": "Yellow_bellied_Flycatcher",
    "43": "Frigatebird",
    "44": "Northern_Fulmar",
    "45": "Gadwall",
    "46": "American_Goldfinch",
    "47": "European_Goldfinch",
    "48": "Boat_tailed_Grackle",
    "49": "Eared_Grebe",
    "50": "Horned_Grebe",
    "51": "Pied_billed_Grebe",
    "52": "Western_Grebe",
    "53": "Blue_Grosbeak",
    "54": "Evening_Grosbeak",
    "55": "Pine_Grosbeak",
    "56": "Rose_breasted_Grosbeak",
    "57": "Pigeon_Guillemot",
    "58": "California_Gull",
    "59": "Glaucous_winged_Gull",
    "60": "Heermann_Gull",
    "61": "Herring_Gull",
    "62": "Ivory_Gull",
    "63": "Ring_billed_Gull",
    "64": "Slaty_backed_Gull",
    "65": "Western_Gull",
    "66": "Anna_Hummingbird",
    "67": "Ruby_throated_Hummingbird",
    "68": "Rufous_Hummingbird",
    "69": "Green_Violetear",
    "70": "Long_tailed_Jaeger",
    "71": "Pomarine_Jaeger",
    "72": "Blue_Jay",
    "73": "Florida_Jay",
    "74": "Green_Jay",
    "75": "Dark_eyed_Junco",
    "76": "Tropical_Kingbird",
    "77": "Gray_Kingbird",
    "78": "Belted_Kingfisher",
    "79": "Green_Kingfisher",
    "80": "Pied_Kingfisher",
    "81": "Ringed_Kingfisher",
    "82": "White_breasted_Kingfisher",
    "83": "Red_legged_Kittiwake",
    "84": "Horned_Lark",
    "85": "Pacific_Loon",
    "86": "Mallard",
    "87": "Western_Meadowlark",
    "88": "Hooded_Merganser",
    "89": "Red_breasted_Merganser",
    "90": "Mockingbird",
    "91": "Nighthawk",
    "92": "Clark_Nutcracker",
    "93": "White_breasted_Nuthatch",
    "94": "Baltimore_Oriole",
    "95": "Hooded_Oriole",
    "96": "Orchard_Oriole",
    "97": "Scott_Oriole",
    "98": "Ovenbird",
    "99": "Brown_Pelican",
    "100": "White_Pelican",
    "101": "Western_Wood_Pewee",
    "102": "Sayornis",
    "103": "American_Pipit",
    "104": "Whip_poor_Will",
    "105": "Horned_Puffin",
    "106": "Common_Raven",
    "107": "White_necked_Raven",
    "108": "American_Redstart",
    "109": "Geococcyx",
    "110": "Loggerhead_Shrike",
    "111": "Great_Grey_Shrike",
    "112": "Baird_Sparrow",
    "113": "Black_throated_Sparrow",
    "114": "Brewer_Sparrow",
    "115": "Chipping_Sparrow",
    "116": "Clay_colored_Sparrow",
    "117": "House_Sparrow",
    "118": "Field_Sparrow",
    "119": "Fox_Sparrow",
    "120": "Grasshopper_Sparrow",
    "121": "Harris_Sparrow",
    "122": "Henslow_Sparrow",
    "123": "Le_Conte_Sparrow",
    "124": "Lincoln_Sparrow",
    "125": "Nelson_Sharp_tailed_Sparrow",
    "126": "Savannah_Sparrow",
    "127": "Seaside_Sparrow",
    "128": "Song_Sparrow",
    "129": "Tree_Sparrow",
    "130": "Vesper_Sparrow",
    "131": "White_crowned_Sparrow",
    "132": "White_throated_Sparrow",
    "133": "Cape_Glossy_Starling",
    "134": "Bank_Swallow",
    "135": "Barn_Swallow",
    "136": "Cliff_Swallow",
    "137": "Tree_Swallow",
    "138": "Scarlet_Tanager",
    "139": "Summer_Tanager",
    "140": "Artic_Tern",
    "141": "Black_Tern",
    "142": "Caspian_Tern",
    "143": "Common_Tern",
    "144": "Elegant_Tern",
    "145": "Forsters_Tern",
    "146": "Least_Tern",
    "147": "Green_tailed_Towhee",
    "148": "Brown_Thrasher",
    "149": "Sage_Thrasher",
    "150": "Black_capped_Vireo",
    "151": "Blue_headed_Vireo",
    "152": "Philadelphia_Vireo",
    "153": "Red_eyed_Vireo",
    "154": "Warbling_Vireo",
    "155": "White_eyed_Vireo",
    "156": "Yellow_throated_Vireo",
    "157": "Bay_breasted_Warbler",
    "158": "Black_and_white_Warbler",
    "159": "Black_throated_Blue_Warbler",
    "160": "Blue_winged_Warbler",
    "161": "Canada_Warbler",
    "162": "Cape_May_Warbler",
    "163": "Cerulean_Warbler",
    "164": "Chestnut_sided_Warbler",
    "165": "Golden_winged_Warbler",
    "166": "Hooded_Warbler",
    "167": "Kentucky_Warbler",
    "168": "Magnolia_Warbler",
    "169": "Mourning_Warbler",
    "170": "Myrtle_Warbler",
    "171": "Nashville_Warbler",
    "172": "Orange_crowned_Warbler",
    "173": "Palm_Warbler",
    "174": "Pine_Warbler",
    "175": "Prairie_Warbler",
    "176": "Prothonotary_Warbler",
    "177": "Swainson_Warbler",
    "178": "Tennessee_Warbler",
    "179": "Wilson_Warbler",
    "180": "Worm_eating_Warbler",
    "181": "Yellow_Warbler",
    "182": "Northern_Waterthrush",
    "183": "Louisiana_Waterthrush",
    "184": "Bohemian_Waxwing",
    "185": "Cedar_Waxwing",
    "186": "American_Three_toed_Woodpecker",
    "187": "Pileated_Woodpecker",
    "188": "Red_bellied_Woodpecker",
    "189": "Red_cockaded_Woodpecker",
    "190": "Red_headed_Woodpecker",
    "191": "Downy_Woodpecker",
    "192": "Bewick_Wren",
    "193": "Cactus_Wren",
    "194": "Carolina_Wren",
    "195": "House_Wren",
    "196": "Marsh_Wren",
    "197": "Rock_Wren",
    "198": "Winter_Wren",
    "199": "Common_Yellowthroat",
}

classnames = [classname_mapping[str(i)] for i in range(200)]
templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
]



import functools
import itertools
import logging
from abc import abstractmethod

import torch
from torch import Tensor, nn
from torch.nn import functional as F
from torchmetrics import Accuracy, MeanMetric
from torchmetrics.classification.accuracy import MulticlassAccuracy
from tqdm.autonotebook import tqdm

from .base_task import BaseTask

log = logging.getLogger(__name__)


class ClassificationTask(BaseTask):
    def __init__(self, task_config):
        super().__init__(task_config)

    @property
    @abstractmethod
    def num_classes(self):
        """
        Returns the number of classes in the dataset.
        """
        pass

    @property
    @abstractmethod
    def test_loader(self):
        """
        Returns a test data loader.
        """
        pass

    @torch.no_grad()
    def evaluate(self, classifier: nn.Module, device=None):
        accuracy: MulticlassAccuracy = Accuracy(
            task="multiclass", num_classes=self.num_classes
        )
        classifier.eval()
        loss_metric = MeanMetric()
        # if fast_dev_run is set, we only evaluate on a batch of the data
        if self.config.get("fast_dev_run", False):
            log.info("Running under fast_dev_run mode, evaluating on a single batch.")
            test_loader = itertools.islice(self.test_loader, 1)
        else:
            test_loader = self.test_loader

        for batch in (
            pbar := tqdm(
                test_loader, desc="Evaluating", leave=False, dynamic_ncols=True
            )
        ):
            inputs, targets = batch
            if device is not None:
                inputs, targets = inputs.to(device), targets.to(device)
            logits: Tensor = classifier(inputs)

            loss = F.cross_entropy(logits, targets)
            loss_metric.update(loss.detach().cpu())
            acc = accuracy(logits.detach().cpu(), targets.detach().cpu())
            pbar.set_postfix(
                {
                    "accuracy": accuracy.compute().item(),
                    "loss": loss_metric.compute().item(),
                }
            )

        acc = accuracy.compute().item()
        loss = loss_metric.compute().item()
        results = {"accuracy": acc, "loss": loss}
        return results



import itertools

import numpy as np


def generate_simplex_grid(n: int, m: int):
    """
    Generate a uniform grid of points on the n-dimensional simplex.

    Examples:

        >>> generate_simplex_grid(3,2)
        array([[0., 0., 1.],
              [0., 1., 0.],
              [1., 0., 0.]], dtype=float32)

        >>> generate_simplex_grid(2,3)
        array([[0. , 1. ],
              [0.5, 0.5],
              [1. , 0. ]], dtype=float32)

    Args:
        n (int): The dimension of the simplex.
        m (int): The number of grid points along each dimension.

    Returns:
        list: A list of n-dimensional vectors representing the grid points.
    """
    m = m - 1
    # Generate all combinations of indices summing up to m
    indices = list(itertools.combinations_with_replacement(range(m + 1), n - 1))
    # Initialize an empty list to store the grid points
    grid_points = []

    # Iterate over each combination of indices
    for idx in indices:
        # Append 0 and m to the indices
        extended_idx = [0] + list(idx) + [m]
        # Compute the vector components by taking the differences between consecutive indices and dividing by m
        point = [(extended_idx[i + 1] - extended_idx[i]) / m for i in range(n)]
        grid_points.append(point)

    return np.array(grid_points, dtype=np.float32)



import logging
from copy import deepcopy
from typing import Optional

from omegaconf import DictConfig, flag_override
from peft import PeftModel
from transformers import AutoModelForSeq2SeqLM

from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils import parse_dtype

log = logging.getLogger(__name__)


def load_lora_model(
    base_model_path: str,
    peft_model_path: str,
    is_trainable: bool = True,
    merge_and_unload: bool = True,
):
    base_model = AutoModelForSeq2SeqLM.from_pretrained(base_model_path)
    model = PeftModel.from_pretrained(
        base_model,
        peft_model_path,
        is_trainable=is_trainable,
    )
    if merge_and_unload:
        model = model.merge_and_unload()
    return model


class Seq2SeqLMPool(BaseModelPool):
    _config_mapping = BaseModelPool._config_mapping | {
        "_tokenizer": "tokenizer",
        "_model_kwargs": "model_kwargs",
    }

    def __init__(
        self,
        models: DictConfig,
        *,
        tokenizer: Optional[DictConfig],
        model_kwargs: Optional[DictConfig] = None,
        **kwargs,
    ):
        super().__init__(models, **kwargs)
        self._tokenizer = tokenizer
        self._model_kwargs = model_kwargs
        if self._model_kwargs is None:
            self._model_kwargs = DictConfig({})
        with flag_override(self._model_kwargs, "allow_objects", True):
            if hasattr(self._model_kwargs, "torch_dtype"):
                self._model_kwargs.torch_dtype = parse_dtype(
                    self._model_kwargs.torch_dtype
                )

    def load_model(self, model_name_or_config: str | DictConfig, *args, **kwargs):
        model_kwargs = deepcopy(self._model_kwargs)
        model_kwargs.update(kwargs)
        return super().load_model(model_name_or_config, *args, **model_kwargs)

    def load_tokenizer(self, *args, **kwargs):
        assert self._tokenizer is not None, "Tokenizer is not defined in the config"
        tokenizer = isinstance(self._tokenizer, *args, **kwargs)
        return tokenizer



import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Union

import hydra.core.global_hydra
from hydra import compose, initialize
from omegaconf import DictConfig, OmegaConf

from fusion_bench.utils import import_object, instantiate
from fusion_bench.utils.instantiate_utils import set_print_function_call

log = logging.getLogger(__name__)


class HydraConfigMixin:
    """
    A mixin for classes that need to be instantiated from a config file.
    """

    @classmethod
    def from_config(
        cls,
        config_name: Union[str, Path],
        overrides: Optional[List[str]] = None,
    ):
        if not hydra.core.global_hydra.GlobalHydra.instance().is_initialized():
            raise RuntimeError("Hydra is not initialized.")
        else:
            cfg = compose(config_name=config_name, overrides=overrides)

        config_groups = config_name.split("/")[:-1]
        for config_group in config_groups:
            cfg = cfg[config_group]

        if "_target_" in cfg:
            # if the config has a _target_ key, check if it is equal to the class name
            target_cls = import_object(cfg["_target_"])
            if target_cls != cls:
                log.warning(
                    f"The _target_ key in the config is {cfg['_target_']}, but the class name is {cls.__name__}."
                )
            with set_print_function_call(False):
                obj = instantiate(cfg)
        else:
            obj = cls(**cfg)

        return obj



from . import register
from .configuration_smile_qwen2 import SmileQwen2Config
from .modeling_smile_qwen2 import (
    SmileQwen2ForCausalLM,
    SmileQwen2ForQuestionAnswering,
    SmileQwen2ForSequenceClassification,
    SmileQwen2Model,
)


import logging
import os
import pickle
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Union

__all__ = ["cache_to_disk"]


log = logging.getLogger(__name__)


def cache_to_disk(file_path: Union[str, Path]) -> Callable:
    """
    A decorator to cache the result of a function to a file. If the file exists,
    the result is loaded from the file. Otherwise, the function is executed and
    the result is saved to the file.

    ## Example usage

    ```python
    @cache_to_disk("path_to_file.pkl")
    def some_function(*args: Any, **kwargs: Any) -> Any:
        # Function implementation
        return "some result"
    ```

    Args:
        file_path (str): The path to the file where the result should be cached.

    Returns:
        Callable: The decorated function.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    assert isinstance(file_path, Path)

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if os.path.exists(file_path):
                log.info(
                    f"Loading cached result of {func.__name__} from {file_path}",
                    stacklevel=2,
                )
                with open(file_path, "rb") as f:
                    return pickle.load(f)
            else:
                result = func(*args, **kwargs)
                file_path.parent.mkdir(parents=True, exist_ok=True)
                with open(file_path, "wb") as f:
                    pickle.dump(result, f)
                return result

        return wrapper

    return decorator



R"""
Examples:

The following command merges eight clip-ViT models using a weighted average approach.
Because `method.normalize` is set to true, the weights are normalized to sum to 1, thus equivalent to simple average.

```bash
fusion_bench \
    method=linear/weighted_average \
    method.normalize=true \
    method.weights=[0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3] \
    modelpool=... \
    taskpool=...
```
"""

import logging
from typing import List, Mapping, Optional, Union  # noqa: F401

import numpy as np
import torch
from typing_extensions import override

from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.state_dict_arithmetic import state_dict_add, state_dict_mul
from fusion_bench.utils.type import StateDictType

log = logging.getLogger(__name__)


class WeightedAverageAlgorithm(BaseAlgorithm, SimpleProfilerMixin):

    _config_mapping = BaseAlgorithm._config_mapping | {
        "normalize": "normalize",
        "weights": "weights",
    }

    def __init__(
        self,
        normalize: bool,
        weights: List[float],
        verbose: bool = True,
        **kwargs,
    ):
        self.normalize = normalize
        self.weights = weights
        self.verbose = verbose
        log.disabled = not self.verbose
        super().__init__(**kwargs)

    @override
    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        """
        Fuses the models in the model pool using a weighted average approach.

        Parameters
            modelpool (ModelPool): The pool of models to be fused.

        Raises
            ValueError: If the number of weights does not match the number of models in the model pool.

        Returns
            forward_model (torch.nn.Module): The resulting model after fusion.
        """
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(modelpool)

        log.info("Fusing models using weighted average.")
        weights = np.asarray(self.weights)
        if len(weights) != len(modelpool.model_names):
            raise ValueError(
                "Number of weights must match the number of models.,"
                f"but got {len(weights)} weights and {len(modelpool.model_names)} models."
                f"weights: {weights}, models: {modelpool.model_names}"
            )
        if self.normalize:
            weights = weights / np.sum(weights)
        if self.verbose:
            print(f"weights: {weights}, normalized: {self.normalize}")

        sd: Optional[StateDictType] = None
        forward_model = None

        for model_name, weight in zip(modelpool.model_names, weights):
            with self.profile("load_model"):
                model = modelpool.load_model(model_name)
            with self.profile("merge weights"):
                if sd is None:
                    sd = state_dict_mul(model.state_dict(keep_vars=True), weight)
                    forward_model = model
                else:
                    sd = state_dict_add(
                        sd, state_dict_mul(model.state_dict(keep_vars=True), weight)
                    )

        forward_model.load_state_dict(sd)
        if self.verbose:
            self.print_profile_summary()
        return forward_model



import logging
from copy import deepcopy

import torch

from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.parameters import count_parameters
from fusion_bench.utils.state_dict_arithmetic import (
    state_dict_mul,
)

from .base_algorithm import SuperposedAlgorithmBase, compare_models

log = logging.getLogger(__name__)


class SuperposedModelSoupAlgorithm(
    SuperposedAlgorithmBase,
):

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(models=modelpool)

        log.info(
            f"Compressing models using superposed model soup.\n"
            f"Models: {modelpool.model_names}"
        )
        models = {}

        # load state dicts
        state_dicts = self._load_state_dicts(modelpool)
        with self.profile("load model"):
            pretrained_model = modelpool.load_model("_pretrained_")
        absorber_state_dict = self._compute_absorber(state_dicts, pretrained_model)
        if absorber_state_dict is not None:
            state_dicts["absorber"] = absorber_state_dict

        with self.profile("compress and retrieve"):
            retrieved_state_dicts, metadata = self._compress_and_retrieve(
                deepcopy(state_dicts), mode="superposed_model_soup"
            )

        with self.profile("retrieve models"):
            for model_idx, model_name in enumerate(modelpool.model_names):
                if self.ms_mode == "average":
                    coefficient = 1 / len(modelpool.model_names)
                    retrieved_state_dict = state_dict_mul(
                        retrieved_state_dicts[model_name], coefficient
                    )
                elif self.ms_mode == "original":
                    retrieved_state_dict = retrieved_state_dicts[model_name]
                else:
                    raise ValueError(f"Unsupported ms_mode: {self.ms_mode}")
                retrieved_model = modelpool.load_model(
                    model_name
                )  # TODO: avoid repeated loading
                # FIXME: for 'all' mode
                for k, v in retrieved_state_dict.items():
                    if v.shape[0] == 1:
                        retrieved_state_dict[k] = v.squeeze(0)
                retrieved_model.load_state_dict(retrieved_state_dict)
                models[model_name] = retrieved_model
                if self.debug >= 1:
                    with self.profile("metadata"):
                        if torch.cuda.is_available():
                            retrieved_state_dicts[model_name] = {
                                k: v.cuda()
                                for k, v in retrieved_state_dicts[model_name].items()
                            }
                            state_dicts[model_name] = {
                                k: v.cuda() for k, v in state_dicts[model_name].items()
                            }
                            retrieved_state_dict = {
                                k: v.cuda() for k, v in retrieved_state_dict.items()
                            }

                        target_layers = metadata["target_layers"]
                        # focus on the superposition retrieval performance on the target layers
                        metadata["superposed_model_retrieval_similarity"][
                            model_name
                        ] = compare_models(
                            retrieved_state_dicts[model_name],
                            state_dicts[model_name],
                            target_layers,
                        )
                        metadata["superposed_model_svd_subspace_similarities"][
                            model_name
                        ] = self._compute_svd_subspace_similarities(
                            state_dicts[model_name],
                            retrieved_state_dicts[model_name],
                            target_layers,
                        )
                        # overall retrieval performance
                        metadata["model_retrieval_similarity"][model_name] = (
                            compare_models(
                                retrieved_state_dict, state_dicts[model_name]
                            )
                        )
                        metadata["model_svd_subspace_similarities"][model_name] = (
                            self._compute_svd_subspace_similarities(
                                state_dicts[model_name], retrieved_state_dict
                            )
                        )
                        # delete the cuda tensors
                        del (
                            retrieved_state_dicts[model_name],
                            state_dicts[model_name],
                            retrieved_state_dict,
                        )
        with self.profile("metadata"):
            if self.debug >= 0:
                (
                    metadata["trainable_param_count_pretrained_model"],
                    metadata["active_param_count_pretrained_model"],
                ) = count_parameters(pretrained_model)
                (
                    metadata["trainable_param_count_retrieved_model"],
                    metadata["active_param_count_retrieved_model"],
                ) = count_parameters(models[modelpool.model_names[0]])
                print(
                    f"Total storage (Gbs) for retrieval and original: {metadata['total_gb_retrieved']} | {metadata['total_gb_original']}"
                )
        self.print_profile_summary()
        return {"models": models, "metadata": metadata}
